
#ifndef NULL
  #define  NULL    ((void *)0)
#endif


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `CyclicRangeRep': Simple-cyclic-distributed range.            //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class CyclicRangeRep : public RangeRep {
public :
  CyclicRangeRep(int _extent, Dimension _dim, const Layout _layout) ;

  RangeRep* prm() ;
  int bas() ;
  int str() ;

  int isScalar() ;

  Subscript* subscr(const int location) ;
  int subscr(const Subscript* i) ;

  RangeRep* subrng(const int _extent, const int _base, const int _stride) ;

  StrideRep* vector(const int dataLen) ;

  void blockInit(Block& b) ;

  void locBlockInit(Block& b) ;

  void allBlocksEnum(Block& b) ;
  int  allBlocksTest(Block& b) ;
  void allBlocksNext(Block& b) ;

protected :
  int numProcs ;  

  Layout layout ;

private :
  void blockParams(Block& b, int i_lo, int i_hi) ;

MEMORY_POOL_DECS
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `CyclicSubRangeRep': subrange of a                            //
//                            simple-cyclic distributed range.         //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class CyclicSubRangeRep : public CyclicRangeRep {
public :
  CyclicSubRangeRep(const int _extent, const int _base, const int _stride,
                    CyclicRangeRep* _parent,
                    Dimension _dim, const Layout _layout) ;
  ~CyclicSubRangeRep() ;

  RangeRep* prm() ;
  int bas() ;
  int str() ;

  Subscript* subscr(const int location) ;
  int subscr(const Subscript* i) ;

  RangeRep* subrng(const int _extent, const int _base, const int _stride) ;

  StrideRep* vector(const int dataLen) ;

  void blockInit(Block& b) ;

  void locBlockInit(Block& b) ;

  void allBlocksEnum(Block& b) ;
  int  allBlocksTest(Block& b) ;
  void allBlocksNext(Block& b) ;

private :
  // Parent primitive range
  CyclicRangeRep* parent ;

  // Alignment parameters
  int base ;
  int stride ;

  int f ;       // GCD of `numProcs' and `stride'.
  int f_s ;     // A solution of:  f_s * stride = f   mod  numProcs
  int i_step ;  // step of (int) x.  Returned by x.blockStep()
  int b_step ;  // step of x.idx()

  void blockParams(Block& b, int i_lo, int i_hi) ;
  void blockRange(int& i_lo, int& i_hi, int _crd) ;

MEMORY_POOL_DECS
} ;


