
// Local subscripts and range representation base class.

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `Pool' : Taken from Stroustrup.                               //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class Pool {
  struct Link {Link* next ; } ;

  const unsigned esize ;
  Link* head ;

  Pool(Pool&) ;

  void grow() ;

  int nalloc ;  // debug
public :
  Pool(unsigned n) ;
  ~Pool() ;

  void* alloc() ;
  void free(void* b) ;
} ;

inline void* Pool :: alloc() {
  if(head == 0) grow() ;
  Link* p = head ;
  head = p->next ;
  nalloc++ ;  // debug
  return p ;
}

inline void Pool :: free(void* b) {
  Link* p = (Link*) b ;
  p->next = head ;
  head = p ;
  nalloc-- ;  // debug
}


#ifdef NO_STATIC_CONSTRUCTORS

#define POOL_NEW void* operator new(size_t) { return pool->alloc() ; }

#define POOL_DELETE void operator delete(void* p) { if(pool) pool->free(p) ; }

#define POOL_STATE static Pool* pool ; friend void RangeInit() ; friend void RangeFinalize() ;

#define MEMORY_POOL_DECL(T) Pool* T :: pool ;

void RangeInit() ;
void RangeFinalize() ;

#else

#define POOL_NEW void* operator new(size_t) { return pool.alloc() ; }

#define POOL_DELETE void operator delete(void* p) { pool.free(p) ; }

#define POOL_STATE static Pool pool ;

#define MEMORY_POOL_DEFN(T) Pool T :: pool(sizeof(T)) ;

#endif


#define MEMORY_POOL_DECS public : POOL_NEW POOL_DELETE protected : POOL_STATE


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// struct `Subscript': The local subscript.                            //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

struct Subscript : public Coord {
  int blk ;
  int sub ;

  int at() ;
  void ta() ;
} ;

inline int Subscript :: at() {
  if(dim) {
    dim.where() ;
    return dim.crd() == crd ;
  }
  else return 1 ;
}

inline void Subscript :: ta() {
  if(dim)
    dim.erewh() ;
}


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `RangeRep': Base class for representation of ranges.          //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class Block ;
class StrideRep ;

class RangeRep : public Subscript {
friend class Range ;
friend class PromoteStrideRep ;          // Accesses `count'
public :
  RangeRep(const int _extent, Dimension _dim) :
      extent(_extent) {dim = _dim ; }
  RangeRep(const Subscript& s) : Subscript(s) {}

  virtual ~RangeRep() {}

  int size() const {return extent ; }

  virtual RangeRep* prm() ;
  virtual int bas() ;
  virtual int str() ;

  virtual int isScalar() ;

  int step(Block& i) ;

  // Subscripting

  virtual Subscript* subscr(const int _i) ;
  virtual int subscr(const Subscript* s) ;

  // Lower level iterators...

  virtual void blockInit(Block& b) ;

  virtual void locBlockInit(Block& b) ;

  virtual void locBlocksEnum(Block& b) ;
  virtual int  locBlocksTest(Block& b) ;
  virtual void locBlocksNext(Block& b) ;

  virtual void allBlocksEnum(Block& b) ;
  virtual int  allBlocksTest(Block& b) ;
  virtual void allBlocksNext(Block& b) ;

  // Child ranges and strides...

  virtual RangeRep* subrng(const int _extent,
                           const int _base, const int _stride) ;

  virtual StrideRep* vector(const int dataLen) ;


  void incr() {count++ ; }
  int decr() {return --count ; }
protected :
  int extent ;
private :
  int count ;

  // Interim support for PCRC Fortran interface:

  int g_lb, g_stride ;

MEMORY_POOL_DECS
} ;

enum Layout {STANDARD, PACKED, ADAPT, SHADOW} ;

// ---------------------------------------------------
//
// Things to do:
//
//     Change name of `RangeRep :: vector'?  Eg, `RangeRep :: arrstr'.
//
//     Later, move `Layout' definition into `Range.h'.

