

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `SFrame': Representation of a sequential array entity.        //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class SFrame {
public :
  //SFrame(char* _data, const int _rank, int* _rngs, int* _strs,
  SFrame(void* _data, const int _rank, int* _rngs, int* _strs,
         const int _typeLen) :
             data(_data), rank(_rank), rngs(_rngs), strs(_strs),
             typeLen(_typeLen) {}

  virtual ~SFrame() {}

  // Inquiry functions
  int rnk() {return rank ; }
  int len() {return typeLen ; }
  //char* dat() {return data ; }
  void* dat() {return data ; }
  int* frm() {return rngs ; }
  int* inj() {return strs ; }
  int rng(const int i) {return rngs [i] ; }
  int str(const int i) {return strs [i] ; }
protected :
  SFrame(const int _rank, const int _typeLen) :
	     rank(_rank), typeLen(_typeLen) {}

  int typeLen ;

  int rank ;
  int* rngs ;
  int* strs ;

  //char* data ;
  void* data ;
private :
  int count ;
friend class SArray ;
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `SSection': Representation of a sequential array section.     //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class Array ;

class SSection : public SFrame {
public :
  //SSection(char* _data,
  SSection(void* _data,
           const int _rank, int* _rngs, int* _strs, const int _typeLen) ;

  SSection(Array parent) ;

  ~SSection() ;
protected :
  SSection(const int _rank, const int _typeLen) ;
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `SData': Representation of a sequential array variable.       //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class SData : public SSection {
public :
  SData(const int _rank, int* _rngs, const int _typeLen) ;

  ~SData() ;
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `SArray': Handle class for members of `SFrame' hierarchy.     //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class SArray {
public :
  SArray() : reprsn(0) {}

  SArray(const SArray& a) : reprsn(a.reprsn) {
    if(reprsn)
      reprsn->count++ ;
  }

  SArray& operator=(const SArray& a) {
    if(a.reprsn)
      a.reprsn->count++ ;
    if(reprsn && --reprsn->count == 0)
      delete reprsn ;
    reprsn = a.reprsn ;
    return *this ;
  }

  ~SArray() {
    if(reprsn && --reprsn->count == 0)
      delete reprsn ;
   }

  // Raw frames
  //SArray(char* data, const int rank, int* rngs, int* strs, const int typeLen,
  SArray(void* data, const int rank, int* rngs, int* strs, const int typeLen,
         const int copy = 0) {
    if(copy)
      reprsn = new SSection(data, rank, rngs, strs, typeLen) ;
    else
      reprsn = new SFrame(data, rank, rngs, strs, typeLen) ;
    reprsn->count = 1 ;
  }

  // Coercion

  SArray(Array parent) ;

  // Data
  SArray(const int rank, int* rngs, const int dataLen) {
    reprsn = new SData(rank, rngs, dataLen) ;
    reprsn->count = 1 ;
  }

  operator void*() {return (void*) reprsn ; }  // used to test if non-null

  int rng(const int d) {return reprsn->rng(d) ; }
  int str(const int d) {return reprsn->str(d) ; }

  int rnk() {return reprsn->rnk() ; }

  int len() {return reprsn->len() ; }
  //char* dat() {return reprsn->dat() ; }
  void* dat() {return reprsn->dat() ; }

  int* frm() {return reprsn->frm() ; }
  int* inj() {return reprsn->inj() ; }
private :
  SFrame* reprsn ;
} ;


// -------------------------------------------------------------
// Things to do
//

