

#ifndef NULL
  #define  NULL    ((void *)0)
#endif


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `StepRangeRep': stepped block-distributed range (C* format).  //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class StepRangeRep : public RangeRep {
public :
  StepRangeRep(int _extent, Dimension _dim, const Layout _layout) ;

  StepRangeRep(int _extent, Dimension _dim,
               const int _blockSize, const int _stepCoord,
               const Layout _layout) ;

  RangeRep* prm() ;
  int bas() ;
  int str() ;

  int isScalar() ;

  Subscript* subscr(const int location) ;
  int subscr(const Subscript* i) ;

  RangeRep* subrng(const int _extent, const int _base, const int _stride) ;

  StrideRep* vector(const int dataLen) ;

  void blockInit(Block& b) ;

  void locBlockInit(Block& b) ;

  void allBlocksEnum(Block& b) ;
  int  allBlocksTest(Block& b) ;
  void allBlocksNext(Block& b) ;

protected :
  int blockSize ;
  int stepCoord ;

  Layout layout ;

  //friend class StepSubRange ;

private :
  void blockParams(Block& b, int i_lo, int i_hi) ;

MEMORY_POOL_DECS
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `StepSubRangeRep': range aligned to a stepped                 //
//                          block-distributed range.                   //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class StepSubRangeRep : public StepRangeRep {
public :
  StepSubRangeRep(const int _extent, const int _base, const int _stride,
                  StepRangeRep* _parent,
                  Dimension _dim, const int _blockSize,
                  const int _stepCoord, const Layout _layout) ;
  ~StepSubRangeRep() ;

  RangeRep* prm() ;
  int bas() ;
  int str() ;

  Subscript* subscr(const int location) ;
  int subscr(const Subscript* i) ;

  RangeRep* subrng(const int _extent, const int _base, const int _stride) ;

  StrideRep* vector(const int dataLen) ;

  void blockInit(Block& b) ;

  void locBlockInit(Block& b) ;

  void allBlocksEnum(Block& b) ;
  void allBlocksNext(Block& b) ;

protected :
  // Parent primitive range
  StepRangeRep* parent ;

  // Alignment parameters
  int base ;
  int stride ;
private :
  void blockParams(Block& b, int i_lo, int i_hi) ;
  void blockRange(int& i_lo, int& i_hi, int _crd) ;

MEMORY_POOL_DECS
} ;

