
// Miscellaneous common operation on range representation objects

#include <stdlib.h>
#include <stdio.h>
#include "globals.h"
#include "Procs.h"
#include "Range.h"


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `Pool' : Taken from Stroustrup.                               //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

inline int max(int x, int y) {return x > y ? x : y ; }

Pool :: Pool(unsigned sz) : esize(max(sizeof(Link*), sz)) {
  head = 0 ;
  nalloc = 0 ;  // debug
}

Pool :: ~Pool() {
  //if(nalloc)
  //  printf("Pool :: ~Pool: %d records remain allocated\n", nalloc) ; // debug
}

void Pool :: grow() {
  const int overhead = 12 ;
  const int chunk_size = 8 * 1024 - overhead ;
  const int nelem = (chunk_size - esize) / esize ;

  char* start = new char [chunk_size] ;
  char* last = &start [(nelem - 1) * esize] ;

  for(char* p = start ; p < last ; p += esize)
    ((Link*) p)->next = (Link*) (p + esize) ;
  ((Link*) last)->next = 0 ;
  head = (Link*) start ;
}


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// Creation and deletion of individual memory pools.                   //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

#ifdef NO_STATIC_CONSTRUCTORS

void RangeInit() {
  // Initialise memory pools.

  CollapsedRangeRep :: pool       = new Pool(sizeof(CollapsedRangeRep))       ;
  CollapsedSubRangeRep :: pool    = new Pool(sizeof(CollapsedSubRangeRep))    ;
  BlockRangeRep :: pool           = new Pool(sizeof(BlockRangeRep))           ;
  PackedBlockRangeRep :: pool     = new Pool(sizeof(PackedBlockRangeRep))     ;
  ExtendedBlockRangeRep :: pool   = new Pool(sizeof(ExtendedBlockRangeRep))   ;
  BlockSubRangeRep :: pool        = new Pool(sizeof(BlockSubRangeRep))        ;
  CyclicRangeRep :: pool          = new Pool(sizeof(CyclicRangeRep))          ;
  CyclicSubRangeRep :: pool       = new Pool(sizeof(CyclicSubRangeRep))       ;
  BlockCyclicRangeRep :: pool     = new Pool(sizeof(BlockCyclicRangeRep))     ;
  BlockCyclicSubRangeRep :: pool  = new Pool(sizeof(BlockCyclicSubRangeRep))  ;
  StepRangeRep :: pool            = new Pool(sizeof(StepRangeRep))            ;
  StepSubRangeRep :: pool         = new Pool(sizeof(StepSubRangeRep))         ;
  RangeRep :: pool                = new Pool(sizeof(RangeRep))                ;

  SimpleStrideRep :: pool         = new Pool(sizeof(SimpleStrideRep))         ;
  MultiblockStrideRep :: pool     = new Pool(sizeof(MultiblockStrideRep))     ;
  PackedStrideRep :: pool         = new Pool(sizeof(PackedStrideRep))         ;
  PackedMultiblockStrideRep :: pool =
                                  new Pool(sizeof(PackedMultiblockStrideRep)) ;
  PromoteStrideRep :: pool        = new Pool(sizeof(PromoteStrideRep))        ;
  ExtendedStrideRep :: pool       = new Pool(sizeof(ExtendedStrideRep))       ;
}

void RangeFinalize() {
  delete CollapsedRangeRep :: pool       ;
  delete CollapsedSubRangeRep :: pool    ;
  delete BlockRangeRep :: pool           ;
  delete PackedBlockRangeRep :: pool     ;
  delete ExtendedBlockRangeRep :: pool   ;
  delete BlockSubRangeRep :: pool        ;
  delete CyclicRangeRep :: pool          ;
  delete CyclicSubRangeRep :: pool       ;
  delete BlockCyclicRangeRep :: pool     ;
  delete BlockCyclicSubRangeRep :: pool  ;
  delete StepRangeRep :: pool            ;
  delete StepSubRangeRep :: pool         ;
  delete RangeRep :: pool                ;

  delete SimpleStrideRep :: pool         ;
  delete MultiblockStrideRep :: pool     ;
  delete PackedStrideRep :: pool         ;
  delete PackedMultiblockStrideRep :: pool ;
  delete PromoteStrideRep :: pool        ;
  delete ExtendedStrideRep :: pool       ;
}

#endif


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `RangeRep': promotion of a subscript.                         //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

// Provide a complete contingent of virtual functions, just to make the
// class concrete.  Only `isScalar' is useful.

//RangeRep* RangeRep :: clone() {return new RangeRep(*this) ; }

RangeRep* RangeRep :: prm() {return 0 ; }
int RangeRep :: bas() {return 0 ; }
int RangeRep :: str() {return 0 ; }

int RangeRep :: isScalar() {return 1 ; }

Subscript* RangeRep :: subscr(const int location) {return 0 ; }
int RangeRep :: subscr(const Subscript* s) {return 0 ; }

RangeRep* RangeRep :: subrng(const int _extent,
			     const int _base, const int _stride) { return 0 ; }

StrideRep* RangeRep :: vector(const int dataLen) {return 0 ; }

void RangeRep :: blockInit(Block& b) {}

void RangeRep :: locBlockInit(Block& b) {}

void RangeRep :: locBlocksEnum(Block& b) {
  if(b.blockCount)
    b.blk = 0 ;
  else
    b.blk = 1 ;
}
int  RangeRep :: locBlocksTest(Block& b) {return b.blk == 0 ; }
void RangeRep :: locBlocksNext(Block& b) {b.blk = 1 ; }

void RangeRep :: allBlocksEnum(Block& b) {}
int  RangeRep :: allBlocksTest(Block& b) {return 0 ; }
void RangeRep :: allBlocksNext(Block& b) {}


// Memory management

MEMORY_POOL_DEFN(RangeRep)

// ---------------------------------------------------
//
// Things to do:
//

