
//   integer n
//   parameter(n = 8)
// 
//   real u(n, n)
// 
//   !hpf$ distribute u (block, block) onto p
//   !hpf$ processors p (2, 2)
// 
//   integer i, j
//   integer iter
// 
//   do i = 1, n
//     do j = 1, n
//       if(i == 1 .or. i == n .or. j == 1 .or. j == n) then
//         u(i, j) = 1.0
//       else
//         u(i, j) = 0.0
//       endif
//     enddo
//   enddo
// 
//   do iter = 1, 5
//     u(2 : n - 1, 2 : n - 1) = &
//                  0.25 * (u(2 : n - 1, 1 : n - 2) + u(2 : n - 1, 3 : n) + &
//                          u(1 : n - 2, 2 : n - 1) + u(3 : n, 2 : n - 1))
//     print *, "iteration =", iter
//     do i = 2, n - 1
//       print "(10f7.3)", u(i, 2 : n - 1)
//     enddo
//   enddo
// 
// end


#include <stdio.h>
#include "ad++.h"
#include "admacros.h"


const int n = 8 ;

int main(int argc, char* argv[]) {
  AdlibInit(argc, argv) ; {

  Procs2 p(2, 2) ;

  ON(p) {
    const int w [2] = {1, 1} ;
    Range x(n, p.dim(0), BLK, SHADOW, 1, 1) ;
    Range y(n, p.dim(1), BLK, SHADOW, 1, 1) ;
    Array2<float> u(x, y) ;

    Index i(x), j(y) ;
    WHERE(i) {
      WHERE(j) {
        if(i == 0 || i == n - 1 || j == 0 || j == n - 1)
          u(i, j) = 1.0 ;
        else
          u(i, j) = 0.0 ;
      } EREWH(j) ;
    } EREWH(i) ;

    Index ic(x.subrng(n - 2, 1)), jc(y.subrng(n - 2, 1)) ;
    Array2<float> uold(x, y) ;

    for(int iter = 1 ; iter <= 5 ; iter++) {
      copy(uold, u) ;
      writeHalo(uold, w, w) ;

      WHERE(ic) {
        WHERE(jc) {
          u(ic, jc) = 0.25 * (uold(ic << 1, jc) + uold(ic >> 1, jc) +
                              uold(ic, jc << 1) + uold(ic, jc >> 1)) ;
        } EREWH(jc) ;
      } EREWH(ic) ;

      gprintf("iteration = %d\n", iter) ;
      aprintf("%7.3f%*N", u, n) ;
    }
  } NO(p) ;

  } AdlibFinalize() ;
}

