

#ifndef NULL
  #define  NULL    ((void *)0)
#endif


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `BlockCyclicRangeRep': Block-cyclically disributed range.     //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class BlockCyclicRangeRep : public RangeRep {
public :
  BlockCyclicRangeRep(int _extent, Dimension _dim, int _blockSize,
                      Layout _layout) ;

  RangeRep* prm() ;
  int bas() ;
  int str() ;

  //int isScalar() ;

  Subscript* subscr(const int location) ;
  int subscr(const Subscript* i) ;

  RangeRep* subrng(const int _extent, const int _base, const int _stride) ;

  StrideRep* vector(const int dataLen) ;

  void blockInit(Block& b) ;
  void locBlockInit(Block& b) ;

  void locBlocksEnum(Block& b) ;
  int locBlocksTest(Block& b) ;
  void locBlocksNext(Block& b) ;

  void allBlocksEnum(Block& b) ;
  int  allBlocksTest(Block& b) ;
  void allBlocksNext(Block& b) ;

protected :
  int blockSize ;
  int layerSize ;

  Layout layout ;

private :
  void blockParams(Block& b, int i_lo, int i_hi) ;

MEMORY_POOL_DECS
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `BlockCyclicSubRangeRep': The most general HPF range,         //
//                      including block-cyclic distribution, and       //
//                      arbitrary strided alignment.                   //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class BlockCyclicSubRangeRep : public BlockCyclicRangeRep {
public :
  BlockCyclicSubRangeRep(const int _extent, const int _base, const int _stride,
                         BlockCyclicRangeRep* _parent,
                         Dimension _dim, int _blockSize, Layout _layout) ;

  ~BlockCyclicSubRangeRep() ;

  RangeRep* prm() ;
  int bas() ;
  int str() ;

  Subscript* subscr(const int location) ;
  int subscr(const Subscript* i) ;

  StrideRep* vector(const int dataLen) ;

  RangeRep* subrng(const int _extent, const int _base, const int _stride) ;

  void blockInit(Block& b) ;
  void locBlockInit(Block& b) ;

  void locBlocksEnum(Block& b) ;
  void locBlocksNext(Block& b) ;

  void allBlocksEnum(Block& b) ;
  void allBlocksNext(Block& b) ;

protected :
  // Parent primitive range
  BlockCyclicRangeRep* parent ;

  // Alignment parameters
  int base ;
  int stride ;
private :
  void blockParams(Block& b, int i_lo, int i_hi) ;
  void blockRange(int& i_lo, int& i_hi, int _crd, int _blk) ;

MEMORY_POOL_DECS
} ;

