
#ifndef NULL
  #define  NULL    ((void *)0)
#endif


class Procs ;
class DimensionRep ;
class Dimension ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//    class `Group'.                                                   //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class Coord ;
class Dimension ;
class DimensionSet ;

class Group {
public :
  Group() {}
  Group(const Group&) ;
  Group(const Procs&) ;

  int lead()  const ;

  //void activeSection() ;  // Replace by `clip'.

  int  on() const ;
  void no() const ;

  inline Group operator/(const Coord& i) const ;
  inline Group& operator/=(const Coord& i) ;

  DimensionSet dims() const ;
  int contains(Group p) const ;

  void clip() ;

  Dimension dim(const int r) const ;
      // Currently used in pcrc :: DIM :: setDistribute, but should be
      // deleted eventually.

protected :
  Procs* procs ;

  int mask ;
  int leadId ;

  void deactivate(const DimensionRep*, const int) ;
  void reactivate(const DimensionRep*, const int) ;

friend class ActiveProcessGroup ;
} ;

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//    class `ActiveProcessGroup': the process group executing this     //
//                                control thread.                      //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class ProcTree ;

class ActiveProcessGroup : public Group {
public :

  // Used in communication library...

  int lead() const ;

  void spanTree(ProcTree*) ;

  // Used from `Procs', `Group', `Dimension', `DimensionSet' respectively...

  int on(Procs*) ;
  void no(Procs*) ;

  int on(const Group*) ;
  void no(const Group*) ;

  void where(DimensionRep*) ;
  void erewh(DimensionRep*) ;

  void where(long _mask) ;
  void erewh(long _mask) ;

  void clip(Group* p) ;

  // Used in `Procs :: Procs', but potentially useful to others...

  int size() const ;

  int locIndex() const ;
private :
  //ActiveProcessGroup() ;
  //ActiveProcessGroup(const ActiveProcessGroup&) ;

  void init(Procs* _procs) ;

  void deactivate(DimensionRep*) ;
  void reactivate(DimensionRep*) ;

  int procId(const int) const ;  // Used in `spanTree' and `Procs :: Procs'

  int onLevel ;

  friend class Procs ;
    // Accesses `ActiveProcessGroup()',
    //          `ActiveProcessGroup(const ActiveProcessGroup&)'
    //          `size()', `locIndex()', `procId()'.

  friend class Group ;

  friend void AdlibInit(int&, char** (&)) ;  // Accesses `init'
} ;


extern ActiveProcessGroup apg ;

inline void Group :: clip() {apg.clip(this) ; }

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//    class `Procs': the process array.                                //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class Procs {
public :
  Procs() ;  // Only used in initialisation
  Procs(const int _rank, int* _n, int* subIds = 0) ;
  ~Procs() ;

private :
  Procs(const Procs& p) ;  // forbid copying.

public :
  int rnk() const {return rank ; }
  Dimension dim(const int i) const ;
  int loc() const {return !outsider ; }

  int on() ;
  void no() ;

  Group operator/(const Coord& i) const ;
private :
  ActiveProcessGroup parent ;
  int rank ;
  Dimension* dims ;
  int* physIds ;
  int outsider ;

  int mask ;

friend class ActiveProcessGroup ;  // accesses `physIds', `parent', `outsider'
friend class Group ;               // accesses `physIds', `mask'
} ;


inline int Group :: lead()  const {return procs->physIds [leadId] ; }


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//    class `DimensionRep': representation class for process array     //
//                          dimension.                                 //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class DimensionRep {
public :
  int size() const {return extent ; }
  int crd() const {return coord ; }
  DimensionRep(const int, const int, const int, Procs*, const int) ;

private :
  int extent ;
  int stride ;
  int coord  ;

  Procs* procs  ;
  int axis ;
  int mask ;

  int inactive ;

friend class ActiveProcessGroup ;  // accesses `active', `extent', `stride', `coord'
friend class Group ;               // accesses `active', `extent', `stride', `axis'
friend class DimensionSet ;        // accesses `mask'
friend class Dimension ;           // accesses `stride'
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
//    class `Dimension': the process array dimension.                  //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class Dimension {
public :
  Dimension() {rep = (DimensionRep*) 0 ; }

  int size() const {return rep->size() ; }
  int crd() const {return rep->crd() ; }

  operator void*() {return rep ; }

  int offset(Coord& i) ;

  void where() {apg.where(rep) ; }
  void erewh() {apg.erewh(rep) ; }

private :
  DimensionRep* rep ;

  void create(const int _extent, const int _coord, const int _stride,
              Procs* _procs, const int _axis) {
    rep = new DimensionRep(_extent, _coord, _stride, _procs, _axis) ;
  }
  void destroy() {delete rep ; }

friend class Procs ;               // accesses `create', `destroy'.
friend class ActiveProcessGroup ;  // accesses `rep'
friend class Group ;               // accesses `rep'
friend class DimensionSet ;        // accesses `rep'

} ;

struct Coord {
  Dimension dim ;
  int crd ;
} ;

inline Dimension Group :: dim(const int r) const {
  // Currently used in pcrc :: DIM :: setDistribute, but should be
  // deleted eventually.

  return procs->dim(r) ;
}

inline int Dimension :: offset(Coord& i) {
  return rep ? rep->stride * i.crd : 0 ;
}

class DimensionSet {
public :
  DimensionSet() : mask(0) {}

  DimensionSet& operator-=(Dimension d) {
    if(d) mask &= ~(d.rep->mask) ;
    return *this ;
  }

  DimensionSet& operator+=(Dimension d) {
    if(d) mask |= d.rep->mask ;
    return *this ;
  }

  DimensionSet operator-(DimensionSet s) const {
    return DimensionSet(mask & ~s.mask) ;
  }

  DimensionSet operator-(Dimension d) const {
    return DimensionSet(mask & ~(d.rep->mask)) ;
  }

  int member(Dimension d) const {return (mask & d.rep->mask) != 0 ; }

  void where() {apg.where(mask) ; }
  void erewh() {apg.erewh(mask) ; }

  int contains(DimensionSet s) const {
    return (~mask & s.mask) == 0 ;
  }

private :
  DimensionSet(const long _mask) : mask(_mask) {}

  long mask ;

  friend class Group ;
  friend class ActiveProcessGroup ;
} ;


inline DimensionSet Group :: dims() const {return DimensionSet(mask) ; }

inline int Group :: contains(Group p) const {

  // Non-strict:
  //   Return value 0 ensures non-containment.
  //   Return value 1 does not ensure containment.

  return procs != p.procs || dims().contains(p.dims()) ;
}

inline int  Group :: on() const {return apg.on(this) ; }
inline void Group :: no() const {apg.no(this) ; }

inline Group Procs :: operator/(const Coord& i) const {
  Group result(*this) ;
  result /= i ;
  return result ;
}

inline Group Group :: operator/(const Coord& i) const {
  Group result(*this) ;
  DimensionRep* dim = i.dim.rep ;
  if(dim)
    result.deactivate(dim, i.crd) ;
  return result ;
}

inline Group& Group :: operator/=(const Coord& i) {
  DimensionRep* dim = i.dim.rep ;
  if(dim)
    deactivate(dim, i.crd) ;
  return *this ;
}


// ---------------------------------------------------------------------
//
// Things to do...
//
//   Add a `procs' pointer to `DimensionSet'?
//
//   Derive `Group' from `DimensionSet'?
//
//   Derive `Procs' from `Group'?  Involves setting `procs' to `this'.
//
//   Need to decide which functions (on `DimensionSet', for example)
//   can accept null dimensions.  I'd incline towards none.  (But the
//   / operator on `Group' is supposed to accept null coordinates...)
//
//   Change `Dimension :: size()' and `Dimension :: crd()' to work for
//   null dimensions, returning 1 and 0 respectively?

