
// Operations on the representation classes for distributed array strides.

#include <stdlib.h>
#include <stdio.h>
#include "globals.h"
#include "Procs.h"
#include "Range.h"


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `SimpleStrideRep'                                             //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

int SimpleStrideRep :: offset(const Subscript* s) {
  return memStep * s->sub ;
}

int SimpleStrideRep :: step(Block& i) {
  return memStep * i.blockStepSub ;
}

int SimpleStrideRep :: volume() {
  return extent ;
}

// Memory management

MEMORY_POOL_DEFN(SimpleStrideRep)


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `MultiblockStrideRep'                                         //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

int MultiblockStrideRep :: offset(const Subscript* s) {
  return blkStep * s->blk + memStep * s->sub ;
}

// Memory management

MEMORY_POOL_DEFN(MultiblockStrideRep)


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `PackedStrideRep'                                             //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

int PackedStrideRep :: offset(const Subscript* s) {
  return memStep * (s->sub / stride) ;
}

int PackedStrideRep :: step(Block& i) {
  return memStep * i.blockStepSub / stride ;
}

// Memory management

MEMORY_POOL_DEFN(PackedStrideRep)

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `PackedMultiblockStrideRep'                                   //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

int PackedMultiblockStrideRep :: offset(const Subscript* s) {
  return blkStep * (s->blk - blkBase) + memStep * (s->sub / stride) ;
}

// Memory management

MEMORY_POOL_DEFN(PackedMultiblockStrideRep)


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `PromoteStrideRep': Stride used to give a distributed         //
//                           ``view'' of a sequential data array.      //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

int PromoteStrideRep :: offset(const Subscript* s) {
  return memStep * parent->subscr(s) ;
}

int PromoteStrideRep :: step(Block& i) {
  return memStep * parent->step(i) ;
}

int PromoteStrideRep :: volume() {
  return 0 ;
}

// Memory management

MEMORY_POOL_DEFN(PromoteStrideRep)


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `ExtendedStrideRep'                                           //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

int ExtendedStrideRep :: offset(const Subscript* s) {
  return memStep * (s->sub + shift) ;
}

MEMORY_POOL_DEFN(ExtendedStrideRep)

