#define Forall_Ref   5
#define Assign_Ref  10
#define Io_Ref      15

#define Proc_Def    20
#define Align_Def   30
#define Temp_Def    40
#define Dist_Def    50

#define Inherit_Def 60

#define NeedDad_Def 70

#define FINDEX_Def  85
#define FHDER_Def   90
#define IHDER_Def   95

#define DUMP_Def   100

#define NOP    	     0
#define BLOCK	    -1
#define CYCLIC	     1

enum Flag { off, on };

// ----------------;

#define MAX_PROCESSOR_DIM 10

///
class Processor_def {
  ///
  int dim;
  ///
  SgSymbol * symb;
  ///
  int s[MAX_PROCESSOR_DIM];
  ///
  int e[MAX_PROCESSOR_DIM];
  ///
  SgSymbol * grp;
  ///
  SgSymbol * shp;
public:
  ///
  Processor_def(void);
  ///
  void setsymbol(SgSymbol *);
  ///
  SgSymbol * getsymbol(void);
  ///
  void add_dimention(int, int);
  ///
  int getDim(void);
  ///
  int getStart(int);
  ///
  int getEnd(int);
  ///
  void setgrp(SgSymbol *);
  ///
  SgSymbol * getgrp(void);
  ///
  void setshp(SgSymbol *);
  ///
  SgSymbol * getshp(void);
  ///
  void show(void);
};

// ---------------

#define MAX_TEMPLATE_DIM 10

///
class Template_def {
  ///
  int dim;
  ///
  SgSymbol * symb;
  ///
  SgType * type;
  ///
  int s[MAX_TEMPLATE_DIM];
  ///
  int e[MAX_TEMPLATE_DIM];
  ///
  int localsize[MAX_TEMPLATE_DIM];
  ///
  int lput; //@ManMemo: largest possible used time;
  SgSymbol * rng;

public:
  ///
  Template_def(void);
  ///
  void setsymbol(SgSymbol *);
  ///
  SgSymbol * getsymbol(void);
  ///
  void settype(SgType *);
  ///
  SgType * gettype(void);
  ///
  void add_dimention(int, int);
  ///
  int getDim(void);
  ///
  int getStart(int);
  ///
  int getEnd(int);
  ///
  void setlocalsize(int, int);
  ///
  int getlocalsize(int);
  ///
  void setlput(int);
  ///
  int getlput();
  ///
  void setrng(SgSymbol *);
  ///
  SgSymbol * getrng(void);
  ///
  void show(void);
};

// ----------------

#define MAX_ALIGN_DIM 10

///
class Alignment_def {
  ///
  int dim;
  ///
  int align_with_dim;
  ///
  SgSymbol * fromsymb;
  ///
  SgSymbol * tosymb;
  ///
  SgExpression * stride[MAX_ALIGN_DIM];
  ///
  SgExpression * offset[MAX_ALIGN_DIM];
  ///
  int align_pair[MAX_ALIGN_DIM];
  ///
  SgSymbol * local;
  ///
  SgSymbol * dad;
  ///
  SgSymbol * tmb;
  ///
  SgSymbol * tmbdad;

public:
  ///
  Alignment_def(void);
  ///
  void setsymbol(SgSymbol *, SgSymbol *);
  ///
  SgSymbol * getfromsymbol(void);
  ///
  SgSymbol * gettosymbol(void);
  ///
  void add_dimention_info(SgExpression *, SgExpression *);
  ///
  int getDim(void);
  ///
  void setAlignWithDim(int);     // dim of align with dimension
  ///
  int getAlignWithDim(void);      
  ///
  SgExpression * getStride(int);
  ///
  SgExpression * getOffset(int);
  ///
  void setAlignPair(int);
  ///
  int getAlignPair(int);
  ///
  void setlocal(SgSymbol *);
  ///
  SgSymbol * getlocal(void);
  ///
  void setdad(SgSymbol *);
  ///
  SgSymbol * getdad(void);
  ///
  void settmb(SgSymbol *);
  ///
  SgSymbol * gettmb(void);
  ///
  void settmbdad(SgSymbol *);
  ///
  SgSymbol * gettmbdad(void);
  
  ///
  void show(void);
};

// ------------;

#define MAX_DISTRIBUTE_DIM 10

///
class Distribute_def {
  ///
  int dim;
  ///
  SgSymbol * fromsymb;
  ///
  SgSymbol * tosymb;
  ///
  int distblock[MAX_DISTRIBUTE_DIM];
public:
  ///
  Distribute_def(void);
  ///
  void setsymbol(SgSymbol *, SgSymbol *);
  ///
  SgSymbol * getfromsymbol(void);
  ///
  SgSymbol * gettosymbol(void);
  ///
  void add_dimention_info(int);
  ///
  int distCode(int);
  ///
  void show(void);
};

// -------------------;

/// 

class Forall_index_def {
  ///
  SgSymbol * indexvar;
  ///
  SgSymbol * lb;
  ///
  SgSymbol * ub;
  ///
  SgSymbol * step;
public:
  ///
  void setsymbol(SgSymbol *);
  ///
  SgSymbol * getsymbol(void);
  ///
  void setlb(SgSymbol *);
  ///
  SgSymbol * getlb(void);
  ///
  void setub(SgSymbol *);
  ///
  SgSymbol * getub(void);
  ///
  void setstep(SgSymbol *);
  ///
  SgSymbol * getstep(void);
};

// ------------;

#define MAX_INDEX_DIM 10

///
class Forall_header_def {
  ///
  int number;
  
  struct {
    SgSymbol * indexvar;
    SgExpression * indextriple[3];
  } index[MAX_INDEX_DIM];

  ///
  SgExpression * condition;

public:
  /// The constructor. 
  Forall_header_def(void);

  ///
  int getindexnumber(void);

  ///
  void setindexsymb(SgSymbol *);
  ///
  SgSymbol * getindexsymb(int);
   
  ///
  void setindextriple(int, SgExpression *);
  ///
  SgExpression * getindextriple(int, int);   

  /// 
  void setcondition(SgExpression *);
  ///
  SgExpression * getcondition(void);

  ///
  int evaluatable_index(int);

   ///
   void show(void);
};

// -------------------------;

enum Position {head, left, right};
enum Communication {IRREG, UNKNOW, NONE, SHIFT, REMAP, BROADCAST};

///
class forall_reference {
  ///
  Position side;

  ///
  SgExpression * expr;
  ///
  int index_level[MAX_ALIGN_DIM][MAX_INDEX_DIM];
  ///
  SgExpression * stride[MAX_ALIGN_DIM][MAX_INDEX_DIM];
  ///
  SgExpression * offset[MAX_ALIGN_DIM][MAX_INDEX_DIM];

  ///
  Communication status;
  ///
  int amount[MAX_ALIGN_DIM];

public:
  ///
  forall_reference (Position);
  ///
  int getposition(void);
  
  ///
  void setExpr(SgExpression *);
  ///
  SgExpression * getexpression(void);
  
  ///
  void setIndex(int, int, int);
  ///
  int getIndex(int, int);

  ///
  void setstride(int, int, SgExpression *);
  ///
  SgExpression * getstride(int, int);

  ///
  void setoffset(int, int, SgExpression *);
  ///
  SgExpression * getoffset(int, int);

  ///
  int scalarDim(int);
  ///
  int tripletDim(int);
  
  ///
  int hasScalarDim(void);

  ///
  int indexNumberOnDim(int);

  ///
  int evaluatable(void);

  ///
  int checkRegular(void);

  ///
  void setCommunication(Communication);
  ///
  Communication getCommunication(void);

  ///
  void setAmount(int, int);
  ///
  int getAmount(int);

  ///
  void show(void);
};

// -----------------------;

///
class assign_reference {
  ///
  Position side;
  ///
  SgExpression * expr;
  
  ///
  Communication status;

public:
  ///
  assign_reference (Position);
  ///
  Position position(void);

  ///
  void setExpression(SgExpression *);
  ///
  SgExpression * expression(void);

  ///
  void setCommunication(Communication);
  ///
  Communication getCommunication(void);

  ///
  void show(void);
};

// -----------------------;

#define MAX_INDEX_DIM 10

///
class Impdo_header_def {
  ///
  int number;
  
  struct { 
    SgSymbol * indexvar;    
    SgExpression * indextriple[3];
  } index[MAX_INDEX_DIM];

public:
  ///
  Impdo_header_def(void);

  ///
  int numberOfIndex(void);
   
  ///
  void setIndexSymbol(SgSymbol *);
  ///
  SgSymbol * indexSymbol(int);
   
  ///
  void setIndexTriplet(int, SgExpression *);
  ///
  SgExpression * getIndexTriplet(int, int);
};

// -----------------------;

///
class io_reference {
  ///
  SgExpression * expr;
  ///
  int index_level[MAX_ALIGN_DIM][MAX_INDEX_DIM];
  ///
  SgExpression * stride[MAX_ALIGN_DIM][MAX_INDEX_DIM];
  ///
  SgExpression * offset[MAX_ALIGN_DIM][MAX_INDEX_DIM];

public:
  ///
  io_reference (void);

  ///
  void setExpression(SgExpression *);
  ///
  SgExpression * expression(void);
  
  ///
  void setIndex(int, int, int);
  ///
  int getIndex(int, int);

  ///
  void setStride(int, int, SgExpression *);

  ///
  SgExpression * getStride(int, int);

  ///
  void setOffset(int, int, SgExpression *);
  ///
  SgExpression * getOffset(int, int);

  ///
  int hasIndexOnDim(int);
  ///
  int indexSequenceOnDim(int);

  ///
  int evaluatable(void);

  ///
  int checkRegular(void);

  ///
  void show(void);
};












