#include <stdio.h>
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#if STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#else
#include <strings.h>
#endif

#include "compilerApi.h"

#include "anls_att.h"

extern char * mixstring;

extern void add_var_dec(SgStatement *, SgType *, SgSymbol *);
extern int getreflevel(SgSymbol * index, SgExpression * expr);
extern SgExpression * getlinerexp(int i, SgSymbol * symb, SgExpression * expr);

void analyse_alignment(SgStatement *);
void show_alignment(SgStatement *);

/*
SgSymbol * symbolinlist(SgExpression *expr, SgExpression * dummy)
{
	for ( ; dummy ; dummy=dummy->rhs() )
	{
		ASSERT (dummy->lhs()->variant() == VAR_REF );
		SgSymbol *sym=dummy->lhs()->symbol();
		if (expr->IsSymbolInExpression(*sym)) return sym;
	}
	return NULL;
}
*/
void add_dimention(Alignment_def * adef, SgExpression * expr)
{
  SgExpression * alignleft = expr->operand(1);
  SgArrayRefExp * alignright = (SgArrayRefExp *)(expr->operand(2)->operand(2));
  
	for (SgExpression *alignl = alignleft ; alignl ; alignl=alignl->rhs())
	{
		SgExpression *source = alignl->lhs();
		ASSERT (source->variant() == VAR_REF);
		SgSymbol   *dummy = source->symbol();
	
		int i=0;
		SgExpression *with = alignright->lhs();
		for ( ; with ; with=with->rhs(),i++ )
		{
			if (with->lhs()->IsSymbolInExpression(*dummy)) break;
		}
		// otherwise collapse
		if (with == NULL) 
		{
			adef->setAlignPair(-1);
    			adef->add_dimention_info(NULL, NULL);
		}
		else 
		{
			SgExpression *node = with->lhs();
/*
		s=getlinerexp(1, index, alignright->subscript(i));
    		o=getlinerexp(0, index, alignright->subscript(i));
*/
    			SgExpression * s=node->lhs()->lhs();
    			SgExpression * o=node->rhs();
			adef->setAlignPair(i);
    			adef->add_dimention_info(s, o);
		}
	}
	adef->setAlignWithDim(alignright->numberOfSubscripts());
}

// deal_Alignment_array add the source and target of alignment to the program header

void deal_Alignment_array(SgStatement * s, SgExpression * expr1, SgExpression * expr2)
{
  Alignment_def * adef =new Alignment_def;

  SgSymbol * srcsymb = expr1->operand(1)->symbol();
  SgSymbol * tarsymb = expr2->operand(1)->operand(2)->operand(2)->symbol();

  // add the alignment source and target;
  adef->setsymbol(srcsymb, tarsymb);

  // the following operand(1) should be an ALIGN_OP;
  add_dimention(adef, expr2->operand(1)); 

  // add align def to the program header and source symbol;
  s->controlParent()->addAttribute(Align_Def, (void *) adef, sizeof(Alignment_def));
  adef->getfromsymbol()->addAttribute(Align_Def, (void *) adef, sizeof(Alignment_def));     
}

void analyse_alignment(SgStatement * s)
{
  deal_Alignment_array(s, s->expr(0), s->expr(2));
}

void show_alignment(SgStatement * s)
{
  for (int i=0; i<s->numberOfAttributes(Align_Def); i++) {
    ((Alignment_def *) ((s->getAttribute(i, Align_Def))->getAttributeData()) ) ->show();
  };
}

//------------------------------------;

void analyse_distribute(SgStatement *);
void show_distribute(SgStatement *);

void add_dist_dimention(Distribute_def * ddef, SgExpression * expr)
{
  int s;

  //printf("----%d\n", expr->id());
  switch (expr->operand(1)->variant()) {
  case STAR_OP:
    s=NOP; 
    break;
  case BLOCK_OP:
    s=BLOCK;
    break;
  case CYCLIC_OP:
    s=CYCLIC;
    break;
  default:
    printf("Wrong definition\n");
    break;    
  };

  ddef->add_dimention_info(s);

  if (expr->operand(2))
    add_dist_dimention(ddef, expr->operand(2));

}

// deal_Distribute_array add the source and target of distribute to the program header

void deal_Distribute_array(SgStatement * s, SgExpression * expr1, SgExpression * expr2)
{
  Distribute_def * ddef = new Distribute_def;

  // add the alignment source and target.
  ddef->setsymbol(expr1->operand(1)->symbol(),expr2->operand(1)->operand(2)->operand(2)->symbol());
  
  // add dimention info, look for the right position.
                     
  add_dist_dimention(ddef, expr2->operand(1)->operand(1)->operand(2));

  s->controlParent()->addAttribute(Dist_Def, (void *) ddef, sizeof(Distribute_def));

  // also add to the from symbol

  ddef->getfromsymbol()->addAttribute(Dist_Def, (void *) ddef, sizeof(Distribute_def));
}

void analyse_distribute(SgStatement * s)
{
  deal_Distribute_array(s, s->expr(0), s->expr(2));
}

void show_distribute(SgStatement * s)
{
  for (int i=0; i<s->numberOfAttributes(Dist_Def); i++) {
    ((Distribute_def *) ((s->getAttribute(i, Dist_Def))->getAttributeData()) ) ->show();
  };
}

//-------------------------;
extern SgType * gettype(int);

void add_align_array(SgStatement * s)
{
  for (int i=0; i<s->numberOfAttributes(Align_Def); i++) {
    Alignment_def * adef = ((Alignment_def *) ((s->getAttribute(i, Align_Def))->getAttributeData()) );

    SgSymbol * arrayname = s->getSymbolInCurrentScope(adef->getfromsymbol()->identifier());

    //also record this symbol;
    //adef->setlocal(arrayname);
    
    SgExpression * array = new SgExpression(ARRAY_REF, NULL, NULL, arrayname);

    SgExpression * tail = array;

    // if is not a template ( then must have inherit attribute )   zheng.gb
    if (adef->gettosymbol()->numberOfAttributes(Temp_Def) == 0)
    {
//	printf("I don't know how to deal with a non-template symbol :%s.  ",arrayname->identifier());
//	ASSERT(adef->gettosymbol()->numberOfAttributes(Temp_Def)  != 0);
	continue;
    }

    int uppervalue=((Template_def*)(adef->gettosymbol()->getAttribute(0, Temp_Def)->getAttributeData()))->getlocalsize(0);
//    for (int j=1; j<adef->getDim(); j++) {
    for (int j=1; j<adef->getAlignWithDim(); j++) {
      uppervalue *= ((Template_def*)(adef->gettosymbol()->getAttribute(0, Temp_Def)->getAttributeData()))->getlocalsize(j);
    };
    ASSERT(uppervalue > 0);

    SgValueExp * c1 = new SgValueExp(0);
    SgValueExp * c2 = new SgValueExp(uppervalue-1);

    SgExpression * range = new SgExpression(RANGE_OP, c1, c2, NULL);
    SgExpression * ddot = new SgExpression(DDOT, range, NULL, NULL);
    SgExpression * ubexpr = new SgExpression(EXPR_LIST, ddot, NULL, NULL);

    tail->setLhs(ubexpr);
    tail=tail->operand(2);

    SgType * atype =adef->getfromsymbol()->type()->internalBaseType();
    SgExpression * arraytype = new SgExpression(TYPE_OP, NULL, NULL, NULL, atype);

    SgStatement * declarray = new SgStatement(VAR_DECL, NULL, NULL, array, arraytype, NULL);
  
    s->childList1(0)->insertStmtBefore(*declarray, *s);
  };
}
//---------------------;

extern void add_var_dec(SgStatement *, SgType *, SgSymbol *);

void add_temp_base(SgStatement * s)
{
}

// ---------------------;

void delete_align_array(SgStatement * s)
{
  if (s->variant()==VAR_DECL) {
    SgDeclarationStatement * declstmt;
    declstmt = (SgDeclarationStatement *) s ;

    SgExpression *  distvar[10];
    int counter=0;

    int i;

    for (i=0; i<declstmt->numberOfVars(); i++) {
      SgSymbol * varsym = declstmt->var(i)->symbol();
      //      printf("The symbol is --- %s\n", varsym->identifier());
      if(varsym->numberOfAttributes(Align_Def)!=0) {
	distvar[counter]=declstmt->var(i);
	counter++;
      }
    }

    for (i=0; i<counter; i++) {
      declstmt->deleteTheVar(* (distvar[i]));
    }

    if (declstmt->numberOfVars()==0)
      declstmt->deleteStmt();
  };
}

//-------------------;
void add_dummy_array(SgStatement * s)
{
  SgProcHedrStmt *  prochedr = (SgProcHedrStmt *) s;
  int org_numberOfParameters = prochedr->numberOfParameters();

  for (int i=0; i<org_numberOfParameters; i++) {
    SgSymbol * parameter = prochedr->parameter(i);
    if (parameter->type()->variant()==T_ARRAY) {
      SgArrayType * arraytype = (SgArrayType *) parameter->type();

      SgArrayRefExp * declist = new SgArrayRefExp ( * parameter);
      if (parameter->numberOfAttributes(Inherit_Def)!=0) {
	for (int dimI = 0; dimI < arraytype->dimension(); dimI++) {
	  SgValueExp * zero = new SgValueExp(0);
	  SgExpression * star = new SgExpression (STAR_RANGE, NULL, NULL, NULL);
	  SgExpression * ddot = new SgExpression (DDOT, zero, star, NULL);
	  
	  declist->addSubscript( *ddot);

	}
      }
      else
	printf("add_dummy_array:not supported yet!\n");

      
      SgVarDeclStmt * decstmt = new SgVarDeclStmt(*declist, * arraytype->baseType());

      s->childList1(0)->insertStmtBefore(*decstmt, *s);
    }
  }

}


// ---------------------;

void delete_dummy_array(SgStatement * s)
{
  if (s->variant()==VAR_DECL) {
    SgDeclarationStatement * declstmt;
    declstmt = (SgDeclarationStatement *) s ;

    SgExpression *  distvar[10];
    int counter=0;

    int i;

    for (i=0; i<declstmt->numberOfVars(); i++) {
      SgSymbol * varsym = declstmt->var(i)->symbol();
      //      printf("The symbol is --- %s\n", varsym->identifier());
      if(varsym->numberOfAttributes(DUMP_Def)!=0) {
	if(varsym->type()->variant()==T_ARRAY) {
	  distvar[counter]=declstmt->var(i);
	  counter++;
	}
      }
    }

    for (i=0; i<counter; i++) {
      declstmt->deleteTheVar(* (distvar[i]));
    }

    if (declstmt->numberOfVars()==0)
      declstmt->deleteStmt();
  };
}

//----------------;

void analyse_inherit(SgStatement * s)
{
  SgInheritStmt * inherit = (SgInheritStmt *) s;

  for (int i=0; i<inherit->numberOfVars(); i++)
    inherit->var(i)->symbol()->addAttribute(Inherit_Def, NULL, 0);
}




