// 				alignroot.C
// 	is_align_root_template()
//	is_align_root_distribute()
//
//  written by Zheng Gengbin

/*
This file contains 2 pass :
1. check in every scope if the align root is template .
   if not so , creat default template the same size as the array ,
   and align the array to this new template .
2. check in every scope if the align root(now be template) have been distributed .
   if not so , create a default processor and default distribute model .
   if the template is not used at all , delete the template from the scope .
these 2 passes are dependent on the align root information collected in info.C
*/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#if STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#else
#include <strings.h>
#endif

#include "compilerApi.h"
#include "pass.h"
#include "ext_func.h"

/***************************************************************************
		is_align_root_template()
****************************************************************************/

SgSymbol  *get_dim_template(SgStatement *hedr_stmt,SgSymbol *sym)
{
	int variant = hedr_stmt->variant();
	ASSERT(variant == PROG_HEDR || variant == PROC_HEDR || variant == FUNC_HEDR );

	SgSymbol *temp = get_temp_variable(hedr_stmt,"TMP");

	SgType *temptype = sym->type()->copyPtr();
	if (temptype->variant() == T_ARRAY)
		((SgArrayType *)temptype)->setBaseType(*SgTypeTemplates());
	else
		temptype = SgTypeTemplates();
	temp->setType(temptype);
	SgExpression *expr_list = ((SgArrayType *)temptype)->getDimList();
	SgExpression *type_op = new SgExpression(TYPE_OP,NULL,NULL,NULL,SgTypeTemplates());

	SgExpression *arrayref=new SgArrayRefExp(*temp,*expr_list);
	SgExpression *ref = new SgExpression(EXPR_LIST,arrayref,NULL,NULL); 

	SgStatement *decl=new SgStatement(TEMPLATE_STMT,NULL,NULL, ref,type_op,NULL);
        hedr_stmt->firstStmtForDeclaration()->insertStmtAfter(*decl);
	
	temp->addAttribute(Temp_Stmt , decl , sizeof(decl));
	return temp;
}

SgSymbol *add_align_to_template(SgStatement *hedr_stmt , SgSymbol *symbol)
{
	int variant = hedr_stmt->variant();
	ASSERT(variant == PROG_HEDR || variant == PROC_HEDR || variant == FUNC_HEDR );

	SgType *type=symbol->type();

	int dim=0;
	if (type->variant() == T_ARRAY)
		dim=((SgArrayType *)type)->dimension();
	
	SgSymbol *temp=get_dim_template(hedr_stmt,symbol);

	SgExpression *expr_list=NULL;
	for (int i=0; i<dim ; i++)
	{
		SgSymbol *dummy = get_temp_variable(hedr_stmt,"DM");
		SgVarRefExp *ref = new SgVarRefExp(dummy);
		expr_list = new SgExpression(EXPR_LIST,ref,expr_list,NULL);
	}
	// scalar not considered
	SgExpression *arrayref = new SgArrayRefExp(*temp,*expr_list);
	SgExpression *align_with = new SgExpression(ALIGN_WITH_OP,NULL,arrayref,NULL);
	SgExpression *align_op = new SgExpression(ALIGN_OP,expr_list->copyPtr(),align_with,NULL);
	expr_list = new SgExpression(EXPR_LIST , align_op ,NULL,NULL);

	SgExpression *varref = new SgVarRefExp(symbol);
	SgExpression *var_list = new SgExpression(EXPR_LIST , varref ,NULL,NULL);

	SgStatement *align_decl=new SgStatement(ALIGN_DECL,NULL,NULL,var_list,NULL,expr_list);

	// insert statement
	if ( (symbol->attributes() &  ALLOCATABLE_BIT) == 0 )
	{
		SgStatement *decl = symbol->declareInStmt(hedr_stmt);
		if (decl==NULL) decl=hedr_stmt->firstStmtForDeclaration();
        	decl->insertStmtAfter(*align_decl,*hedr_stmt);
	}
	else {
	}

	// add attributes
	temp->addAttribute(Align_Son , symbol , sizeof(symbol));
        symbol->addAttribute(Align_Parent , temp , sizeof(temp));
        symbol->addAttribute(Align_Stmt , align_decl , sizeof(align_decl));
	
	return temp;
}

// make sure if all the symbol align to a Template   ( if INHERIT , then no )
// dependence the  Align_Root collect !!
// must be   done before ALIGN_COLLAPSE_PASS
void is_align_root_template(SgStatement *hedr_stmt)
{
	for (int i=0 ; i< hedr_stmt->numberOfAttributes(Align_Root) ; i++)
        {
                SgSymbol *sym=(SgSymbol*)hedr_stmt->attributeValue(i,Align_Root);
		if (sym->type()->variant() == T_TEMPLATE) continue;
		if (((SgArrayType *)sym->type())->baseType()->variant() == T_TEMPLATE) continue;
		// if have INHERIT attribute
		if (sym->attributes() & INHERIT_BIT) continue;

printf("Align Root: %s is not a template \n",sym->identifier());

		SgSymbol *temp=add_align_to_template(hedr_stmt,sym);

		// modified the distribute statement concerned about the sym
		if (sym->numberOfAttributes(Dist_Stmt) > 0)
		{
			SgStatement *dist_stmt = (SgStatement *)sym->attributeValue(0,Dist_Stmt);
			dist_stmt->expr(0)->lhs()->setSymbol(temp);
printf("Change the distribute %s to template %s.\n",sym->identifier(),temp->identifier());

			// modified Dist_Stmt attribute .
			sym->deleteAttribute(0,Dist_Stmt);
			temp->addAttribute(Dist_Stmt , dist_stmt ,sizeof(dist_stmt) );
		}
        }
}


/***************************************************************************
		is_align_root_distribute()
****************************************************************************/

void distribute_default(SgStatement *hedr_stmt , SgSymbol *symbol)
{
	int variant = hedr_stmt->variant();
	ASSERT(variant == PROG_HEDR || variant == PROC_HEDR || variant == FUNC_HEDR );

	SgType *type=symbol->type();

	SgSymbol *proc;
        int dim=0;
        if (type->variant() == T_ARRAY)
                dim=((SgArrayType *)type)->dimension();

	int d = dim;
	while ((proc=get_dim_processor(hedr_stmt,d))==NULL && d>0 ) d--;

	SgExpression *expr_list=NULL;
        for (int i=1; i<=dim ; i++)
        {
                SgExpression *ref ;
		if (i>d) 
			ref= new SgExpression(STAR_OP,NULL,NULL,NULL);
		else
			ref= new SgExpression(BLOCK_OP,NULL,NULL,NULL);
                expr_list = new SgExpression(EXPR_LIST,ref,expr_list,NULL);
        }
	SgExpression *dist_format=new SgExpression(DIST_FORMAT_OP,NULL,expr_list,NULL);
	SgExpression *varref = new SgVarRefExp(proc);
	SgExpression *dist_target=new SgExpression(DIST_TARGET_OP,NULL,varref,NULL);
	SgExpression *dist_op=new SgExpression(DISTRIBUTE_OP,dist_format,dist_target,NULL);
	expr_list = new SgExpression(EXPR_LIST,dist_op,NULL,NULL);

	varref = new SgVarRefExp(symbol);
	SgExpression *var_list = new SgExpression(EXPR_LIST,varref,NULL,NULL);

	SgStatement *dist_decl=new SgStatement(DISTRIBUTE_DECL,NULL,NULL,var_list,NULL,expr_list);

	// insert statement
	SgStatement *decl = symbol->declareInStmt(hedr_stmt);
	if (decl==NULL) decl=hedr_stmt->firstStmtForDeclaration();
        decl->insertStmtAfter(*dist_decl,*hedr_stmt);
}

//  must be done last
// if has INHERIT attr , donot do
void is_align_root_distribute(SgStatement *hedr_stmt)
{
	for (int i=0 ; i< hedr_stmt->numberOfAttributes(Align_Root) ; i++)
        {
                SgSymbol *sym=(SgSymbol*)hedr_stmt->attributeValue(i,Align_Root);
		if (sym->numberOfAttributes(Dist_Stmt) > 0) continue;

		SgType *atype =sym->type();
                if (atype->variant() == T_ARRAY ) atype=atype->baseType();
 	        if (atype->variant() == T_TEMPLATE)   // no align son 
			if (sym->numberOfAttributes(Align_Son)==0) 
			{
				// then this template seems useless
printf("Delete an unused template : %s\n", sym->identifier());
				SgVarDeclStmt *stmt = (SgVarDeclStmt*)sym->attributeValue(0,Temp_Stmt);
				stmt->deleteTheSymbol(*sym);
				continue;
			}
		// if have INHERIT attribute
		if (sym->attributes() & INHERIT_BIT) continue;

		// make a default distribute
printf("Align Root: %s is not distributed \n",sym->identifier());
		distribute_default(hedr_stmt,sym);
	}
}


