#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "compilerApi.h"
#include "pass.h"

void display_align_root(SgStatement *hedr_stmt)
{
	printf("Align Root: \n");
	for (int i=0 ; i< hedr_stmt->numberOfAttributes(Align_Root) ; i++)
	{
		SgSymbol *sym=(SgSymbol*)hedr_stmt->attributeValue(i,Align_Root);
		printf("root: %s\n",sym->identifier());
	}
	printf("Align Root end \n");
}

void display_align_tree(SgStatement *hedr_stmt)
{
	SgSymbol * symbol = hedr_stmt->symbol()->next();
        SgSymbol * sym_son , *sym_par;

 	printf("Align TREE: \n");
        while ((symbol!=NULL)) {
                if (symbol->scope()==hedr_stmt)
                        break;
                symbol=symbol->next();
        }
        if (symbol==NULL) return ;

        for( SgSymbol *symb=symbol ; symb; symb=symb->nextInScope()) {
		if (symb->scope() == NULL ) continue;
                if (symb->numberOfAttributes(Align_Son) >0 )
                {
			printf("++%s %d-%d: ",symb->identifier(),symb->numberOfAttributes(Align_Son),symb->numberOfAttributes(Align_Parent));
			for (int i =0; i<symb->numberOfAttributes(Align_Son) ; i++)
			{
                        	sym_son=(SgSymbol *)symb->attributeValue(i,Align_Son);
				printf("-- %s ",sym_son->identifier());
			}
			printf("\n");
	        }
        }
        printf("Align TERR end \n");
}

// collect variable which is root of the tree
void collect_align_root(SgStatement *hedr_stmt)
{
	SgSymbol * symb = hedr_stmt->symbol()->next();
	SgType    *atype;

	// clear all attributes about Align_Root
	hedr_stmt->deleteAllAttribute(Align_Root);

	while ((symb!=NULL)) {
    		if (symb->scope()==hedr_stmt)
      			break;
    		symb=symb->next();
  	}
	if (symb==NULL) return ;

	for( ; symb; symb=symb->nextInScope()) {
		if (symb->scope() == NULL ) continue;
		if (symb->scope() != hedr_stmt ) continue;
		if (symb->type() == NULL) continue;
		atype =symb->type();
		// scalar not considered
		if (atype->variant() != T_ARRAY ) continue;
		atype=atype->baseType();
		if (atype->variant() == T_PROCESSORS) continue;
		if (symb->numberOfAttributes(Align_Parent) == 0)
			hedr_stmt->addAttribute(Align_Root,symb,sizeof(symb) );
    	}
}

//   fill the symbol with Align_Parent Align_Son
void  collect_align_attributes(SgStatement *s)
{
	SgExpression *align_op,*align_with_op;
	SgSymbol     *align_par,*align_son;
	SgStatement *hedr_stmt = s->getScopeForDeclare();

	align_op = s->expr(2)->lhs();
        align_with_op=align_op->rhs();

        align_par = align_op->rhs()->rhs()->symbol();
        align_son= s->expr(0)->lhs()->symbol();
	
	if (align_par == align_son )
	{
		printf("%s is align to itself !\n",align_par->identifier() );
		s->deleteStmt();
		return;
	}

	align_par->addAttribute(Align_Son , align_son , sizeof(align_son));
	align_son->addAttribute(Align_Parent , align_par , sizeof(align_par));

	align_son->addAttribute(Align_Stmt , s , sizeof(s));

	// if *( ) presents 
	if (align_with_op->lhs())      
		align_son->addAttribute(Align_Star, NULL,0);
}

//   fill the symbol with Dist_Stmt         must be done only once 
void  collect_dist_attributes(SgStatement *s)
{
        SgSymbol *distributee = s->expr(0)->lhs()->symbol();

        distributee->addAttribute(Dist_Stmt , s , sizeof(s));
}

//   fill the symbol with Temp_Stmt         must be done only once 
void  collect_temp_attributes(SgStatement *s)
{
	for ( SgExpression *expr_list = s->expr(0) ; expr_list ; expr_list=expr_list->rhs())
		expr_list->lhs()->symbol()->addAttribute(Temp_Stmt , s , sizeof(s));
}

