#include <stdio.h>

#include "compilerApi.h"

#include "anls_att.h"

extern char * mixstring;

extern char * stackNameOfType(SgType *);

extern char * allocNameOfType(SgType *);
extern char * freeNameOfType(SgType *);

#define DEMO 0

assign_reference::assign_reference(Position s)
{
  side =s;
  status = NONE;
}

Position assign_reference::position()
{
  return side;
}

void assign_reference::setExpression(SgExpression * e)
{
  expr=e;
}

SgExpression * assign_reference::expression()
{
  return expr;
}

void assign_reference::setCommunication(Communication c)
{
  status=c;
}

Communication assign_reference::getCommunication()
{
  return status;
}

void assign_reference::show()
{
}

//-----------------------------------------------;


void add_assign_reference(SgAssignStmt * assign, SgExpression * exp, Position side)
{
  switch (exp->variant()) {

  case ADD_OP:
  case MINUS_OP:
  case MULT_OP:
  case DIV_OP:
  case MOD_OP: {
    add_assign_reference(assign, exp->operand(1), side);
    add_assign_reference(assign, exp->operand(2), side);
    break;
  }

  case VAR_REF:
  case ARRAY_REF: {
    assign_reference * aref = new assign_reference(side);
    aref->setExpression(exp);
    assign->addAttribute(Assign_Ref,(void *)aref, sizeof(assign_reference));
    break;
  }

  case INT_VAL:
  case FLOAT_VAL:
  case STRING_VAL: {
    // do nothing;
    break;
  }
  
  case INTRIN_FUNC_CALL: {
    // here is a junck code, just for demo;
    assign->addAttribute(DEMO, NULL, 0);
    break;
  }

  default:
    printf("In add_assign_reference, unexpected expression on line %d, id: %d, variant: %d\n",assign->lineNumber(), exp->id(),exp->variant());
    break;

  }
  
}

void detect_assign_reference(SgAssignStmt * assign)
{
  
  assign_reference * leftref = (assign_reference *) ((assign->getAttribute(0, Assign_Ref))->getAttributeData());

  if (leftref->expression()->variant()==VAR_REF) {
    for (int i=1; i<assign->numberOfAttributes(Assign_Ref); i++) {
      assign_reference * aref = (assign_reference *) ((assign->getAttribute(i, Assign_Ref))->getAttributeData());
      aref->setCommunication(REMAP);
    }
  }
  else { // this is an array;
    if (leftref->expression()->symbol()->numberOfAttributes(Align_Def)==0) {
      if (leftref->expression()->symbol()->numberOfAttributes(Inherit_Def)==0) {
	for (int i=1; i<assign->numberOfAttributes(Assign_Ref); i++) {
	  assign_reference * aref = (assign_reference *) ((assign->getAttribute(i, Assign_Ref))->getAttributeData());
	  aref->setCommunication(REMAP);
	}
      }
      else {            // after stdpass , I think this will never happen .
	printf("detect_assign_reference:not finished yet!\n");
      }
    }
    else {
      printf("detect_assign_reference:not finished yet!\n");
    }
  }
}

void analyse_assign(SgStatement * s)
{
  SgAssignStmt * assign = (SgAssignStmt *) s;

  add_assign_reference(assign, assign->lhs(), left);

  add_assign_reference(assign, assign->rhs(), right);

  detect_assign_reference(assign);

  assign_reference * leftref = (assign_reference *) ((assign->getAttribute(0, Assign_Ref))->getAttributeData());

  if ((leftref->expression()->variant()==VAR_REF) ||
      ((leftref->expression()->symbol()->numberOfAttributes(Align_Def)==0)&&
       (leftref->expression()->symbol()->numberOfAttributes(Inherit_Def)==0))) {

    for (int i=1; i<assign->numberOfAttributes(Assign_Ref); i++) {
      assign_reference * aref = (assign_reference *) ((assign->getAttribute(i, Assign_Ref))->getAttributeData());
     
      if (aref->getCommunication()==REMAP) {
	if (leftref->expression()->symbol()->numberOfAttributes(NeedDad_Def)==0) {
	  leftref->expression()->symbol()->addAttribute(NeedDad_Def, NULL, 0);
	  break;
	}
      }
    }
  }

}

void create_section(SgAssignStmt * assign, int counter)
{
  assign_reference * aref = (assign_reference *) ((assign->getAttribute(counter, Assign_Ref))->getAttributeData());
  
  SgArrayRefExp * array = (SgArrayRefExp *) aref->expression();
  SgSymbol * arraysymb  = aref->expression()->symbol();

  {
    SgSymbol * fun = assign->getSymbolInCurrentScope(mixstring, "new_array_section", NULL);
    SgFunctionRefExp * funExp = new SgFunctionRefExp (*fun);

    SgValueExp * zero = new SgValueExp(0);

    funExp->addArg(*zero);

    SgSymbol * dad = assign->getSymbolInCurrentScope(mixstring, "dad", arraysymb->identifier());
    SgExpression * dadexp = new SgExpression(VAR_REF, NULL, NULL, dad);

    funExp->addArg(*dadexp);

    SgSymbol * sym=assign->getSymbolInCurrentScope(mixstring, "sdd", NULL, counter);
    SgExpression * dadsecExp = new SgExpression (VAR_REF, NULL, NULL, sym);
    
    SgAssignStmt * assignStmt = new SgAssignStmt (*dadsecExp, *funExp);
    
    assign->insertStmtBefore(*assignStmt, * assign->controlParent());
  }

  for (int i=0; i<array->numberOfSubscripts(); i++) {
    if (1) {
      SgSymbol * symb= assign->getSymbolInCurrentScope(mixstring, "set_array_scalar", NULL);
      SgCallStmt * funcall= new SgCallStmt(*symb);
      {
	SgSymbol * symb = assign->getSymbolInCurrentScope(mixstring, "sdd", NULL, counter);
	SgExpression * dadsecexp = new SgExpression(VAR_REF, NULL, NULL, symb);

	funcall->addArg(*dadsecexp);
      }
      {
	SgExpression * scalar = ( (SgArrayRefExp *) aref->expression() ) ->subscript(i);
	funcall->addArg(*scalar);
      }
      {
        SgSymbol * dad = assign->getSymbolInCurrentScope(mixstring, "dad", arraysymb->identifier());
	SgExpression * dadexp = new SgExpression(VAR_REF, NULL, NULL, dad);
	
	funcall->addArg(*dadexp);
      };
      {
	SgValueExp * dim = new SgValueExp (i+1);
	funcall->addArg(*dim);
      }
      assign->insertStmtBefore(*funcall, *assign->controlParent());

    }
    else {
      printf("It should be a scalar on this dim\n");
    }

  }

  // call delete_array;
  {
    SgSymbol * symb = assign->getSymbolInCurrentScope(mixstring, "delete_array", NULL);
    SgCallStmt * funcall = new SgCallStmt (*symb);

    SgSymbol * base = assign->getSymbolInCurrentScope(mixstring, "sdd", NULL, counter);
    SgExpression * baseexpr=new SgExpression(VAR_REF, NULL, NULL, base);
    funcall->addArg(*baseexpr);
    
    assign->insertStmtAfter(*funcall, *assign->controlParent());
  }
}

void create_new(SgAssignStmt * assign, int counter)
{
  assign_reference * leftref = (assign_reference *) ((assign->getAttribute(0, Assign_Ref))->getAttributeData());
  SgSymbol * leftsymb = leftref->expression()->symbol();

  assign_reference * aref = (assign_reference *) ((assign->getAttribute(counter, Assign_Ref))->getAttributeData());
  
  SgArrayRefExp * array = (SgArrayRefExp *) aref->expression();
  SgSymbol * arraysymb  = aref->expression()->symbol();

  {
    SgSymbol * symb = assign->getSymbolInCurrentScope(mixstring, "new_array_copy", NULL);
    SgFunctionRefExp * funcall = new SgFunctionRefExp (*symb);

    SgSymbol * stack = assign->getSymbolInCurrentScope(mixstring, stackNameOfType(arraysymb->type()), NULL); 
    SgExpression * stackExp = new SgExpression(VAR_REF, NULL, NULL, stack);
    funcall->addArg(*stackExp);    
						       
    SgSymbol * secDad = assign->getSymbolInCurrentScope(mixstring, "dad", leftsymb->identifier());
    SgExpression * secDadExp=new SgExpression(VAR_REF, NULL, NULL, secDad);
    funcall->addArg(*secDadExp);

    SgSymbol * dadBase = assign->getSymbolInCurrentScope(mixstring, "tdd", NULL, counter);
    SgExpression * dadBaseExp=new SgExpression(VAR_REF, NULL, NULL, dadBase);

    SgAssignStmt * assignStmt = new SgAssignStmt(*dadBaseExp, *funcall);
  
    assign->insertStmtBefore(*assignStmt, *assign->controlParent());
  }

  // allocate memory;
  {
    SgSymbol * base = assign->getSymbolInCurrentScope(mixstring, "tbs", NULL, counter);
    SgExpression * baseexp=new SgExpression(VAR_REF, NULL, NULL, base);

    SgSymbol * franame = assign->getSymbolInCurrentScope(mixstring, allocNameOfType(arraysymb->type()), NULL);
    SgFunctionRefExp * rafuncexpr = new SgFunctionRefExp (*franame);

    SgSymbol * fsizename = assign->getSymbolInCurrentScope(mixstring, "size", NULL);
    SgFunctionRefExp * sizefuncexpr = new SgFunctionRefExp (*fsizename);

    SgSymbol * dadbase = assign->getSymbolInCurrentScope(mixstring, "tdd", NULL, counter);
    SgExpression * dadbaseexp=new SgExpression(VAR_REF, NULL, NULL, dadbase);
    sizefuncexpr->addArg(*dadbaseexp);

    rafuncexpr->addArg(*sizefuncexpr);

    SgStatement * alloc = new SgAssignStmt(*baseexp, *rafuncexpr);

    assign->insertStmtBefore(*alloc, *assign->controlParent());
  }

  // reset base;
  {
    SgSymbol * symb = assign->getSymbolInCurrentScope(mixstring, "reset_array_base", NULL);
    SgCallStmt * funcall = new SgCallStmt (*symb);

    SgSymbol * base = assign->getSymbolInCurrentScope(mixstring, "tdd", NULL, counter);
    SgExpression * baseexpr=new SgExpression(VAR_REF, NULL, NULL, base);
    funcall->addArg(*baseexpr);

    SgSymbol * rs= assign->getSymbolInCurrentScope(mixstring, stackNameOfType(arraysymb->type()), NULL);
    SgArrayRefExp * stackarray = new SgArrayRefExp(*rs);

    SgSymbol * basesym = assign->getSymbolInCurrentScope(mixstring, "tbs", NULL,counter);
    SgExpression * baseexp= new SgExpression(VAR_REF, NULL, NULL, basesym);

    stackarray->addSubscript(*baseexp);
    funcall->addArg(*stackarray);

    assign->insertStmtBefore(*funcall, *assign->controlParent());
  }

  // call delete_array;
  {
    SgSymbol * symb = assign->getSymbolInCurrentScope(mixstring, "delete_array", NULL);
    SgCallStmt * funcall = new SgCallStmt (*symb);

    SgSymbol * base = assign->getSymbolInCurrentScope(mixstring, "tdd", NULL, counter);
    SgExpression * baseexpr=new SgExpression(VAR_REF, NULL, NULL, base);
    funcall->addArg(*baseexpr);
    
    assign->insertStmtAfter(*funcall, *assign->controlParent());
  }

  // call real free;
  {
    SgSymbol * symb = assign->getSymbolInCurrentScope(mixstring, freeNameOfType(arraysymb->type()), NULL);
    SgCallStmt * funcall = new SgCallStmt (*symb);
    
    SgSymbol * base = assign->getSymbolInCurrentScope(mixstring, "tbs", NULL, counter);
    SgExpression * baseexpr=new SgExpression(VAR_REF, NULL, NULL, base);
    
    funcall->addArg(*baseexpr);
    
    assign->insertStmtAfter(*funcall, *assign->controlParent());
  }

}

void add_remap(SgAssignStmt * assign, int counter)
{
  SgSymbol * symb;
  
  symb=assign->getSymbolInCurrentScope(mixstring, "remap", NULL);
  
  SgCallStmt * funcall= new SgCallStmt(*symb);

  symb=assign->getSymbolInCurrentScope(mixstring, "tdd", NULL, counter);
  SgExpression * dadexp = new SgExpression(VAR_REF, NULL, NULL, symb);
  funcall->addArg(*dadexp);

  symb=assign->getSymbolInCurrentScope(mixstring, "sdd", NULL , counter);
  SgExpression * dadsecexp = new SgExpression(VAR_REF, NULL, NULL, symb);
  funcall->addArg(*dadsecexp);

  assign->insertStmtBefore(*funcall, *assign->controlParent());
}

void change_assign_reference(SgAssignStmt * assign, int counter)
{
  assign_reference * aref = (assign_reference *) ((assign->getAttribute(counter, Assign_Ref))->getAttributeData());

  SgArrayRefExp * array = (SgArrayRefExp *) aref->expression();
  SgSymbol * arraysymb  = aref->expression()->symbol();

  SgSymbol * stack = assign->getSymbolInCurrentScope(mixstring, stackNameOfType(arraysymb->type()), NULL); 
  SgArrayRefExp * stackExp = new SgArrayRefExp(* stack);
  
  SgSymbol * base = assign->getSymbolInCurrentScope(mixstring, "tbs", NULL, counter);
  SgExpression * baseexp=new SgExpression(VAR_REF, NULL, NULL, base);
  stackExp->addSubscript(*baseexp);

  array->replace(*stackExp);
}

void trans_assign_reference(SgAssignStmt * assign, int counter)
{
  assign_reference * aref = (assign_reference *) ((assign->getAttribute(counter, Assign_Ref))->getAttributeData());

  if (counter==0) {
    printf("No need now\n");
  }
  else {
    if (aref->getCommunication()==REMAP) {
      create_section(assign, counter);
      create_new(assign, counter);
      add_remap(assign, counter);
      change_assign_reference(assign, counter);
    };
  };
}


void trans_assign(SgStatement * s)
{
  SgAssignStmt * assign = (SgAssignStmt *) s;

  if (assign->numberOfAttributes(DEMO)!=0) {
    // printf("something for demo cshift\n");

    SgSymbol * shift = s->getSymbolInCurrentScope(mixstring, "shift", NULL);
    SgCallStmt *funCall=new SgCallStmt(*shift);


    SgSymbol * ddad = s->getSymbolInCurrentScope(mixstring,"dad", assign->lhs()->symbol()->identifier());
    SgExpression * ddadExp=new SgExpression(VAR_REF, NULL, NULL, ddad);

    funCall->addArg(* ddadExp);

    SgFunctionRefExp * funRef = (SgFunctionRefExp *) assign->rhs();

    SgSymbol * sdad = s->getSymbolInCurrentScope(mixstring,"dad", funRef->arg(0)->symbol()->identifier());
    SgExpression * sdadExp=new SgExpression(VAR_REF, NULL, NULL, sdad);

    funCall->addArg(* sdadExp);
   
    funCall->addArg(* funRef->arg(1));
    funCall->addArg(* funRef->arg(2));


    s->insertStmtBefore(*funCall, *s->controlParent());

    s->deleteStmt();
    return;
  }

  if (s->hasLabel()) {
    SgLabel * label = s->label();

    SgStatement * continueStmt = new SgStatement(CONTINUE_NODE);

    continueStmt->setLabel(*label);

    s->insertStmtBefore(*continueStmt, *s->controlParent());

    s->setLabel(NULL);
  }

  int i;

  for (i=1; i<assign->numberOfAttributes(Assign_Ref); i++) {
    trans_assign_reference(assign, i);
  }
}














