#include <stdio.h>
#include <stdlib.h>

#include "compilerApi.h"

#include "anls_att.h"

extern char * mixstring;

void analyse_template(SgStatement *);
void show_template(SgStatement *);
void add_template_declararion(SgStatement *);

void fill_template_table(Template_def * tdef, SgExpression * expr)
{
  //printf("Current Expr: %d\n", expr->id());
  switch(expr->variant()) {
  case EXPR_LIST:
    fill_template_table(tdef, expr->lhs() );
    if (expr->operand(2))
      fill_template_table(tdef, expr->operand(2));
    break;

  case DDOT:
    ASSERT(expr->lhs());
    fill_template_table(tdef, expr->lhs());
    break;

  case RANGE_OP:
    {
      int start, end;

      if (expr->operand(1)) {
	if (expr->operand(1)->isInteger())
	  start=expr->operand(1)->valueInteger();
	else if (expr->operand(1)->symbol()->variant()==CONST_NAME) { // should be a parameter value ; 
	  //printf ("The id is %d\n", ((SgConstantSymb *) expr->operand(2)->symbol())->constantValue()->id() );
	  if (((SgConstantSymb *) expr->operand(1)->symbol())->constantValue()->isInteger())
	    start =((SgConstantSymb *) expr->operand(2)->symbol())->constantValue()->valueInteger();
	  else
	    printf("It should be an integer\n");
	}
	else
	  printf("Unknow template boundaries.\n");
      }      
      else
	start=1;

      if (expr->operand(2)->isInteger())
	end=expr->operand(2)->valueInteger();
      else if (expr->operand(2)->symbol()->variant()==CONST_NAME) {  // should be a parameter value ; 
	//printf ("The id is %d\n", ((SgConstantSymb *) expr->operand(2)->symbol())->constantValue()->id() );
	if (((SgConstantSymb *) expr->operand(2)->symbol())->constantValue()->isInteger())
	  end =((SgConstantSymb *) expr->operand(2)->symbol())->constantValue()->valueInteger();
	else
	  printf("It should be an integer\n");
      }
      else 
	printf("Unknow template boundaries.\n");

      tdef->add_dimention(start, end);
    };
    break;

  default:
    printf("Wrong array index in declaration: %d\n", expr->id());
  };
}

void deal_template_array(SgStatement * s, SgExpression * expr)
{
  switch(expr->variant()) {
  case EXPR_LIST:
    deal_template_array(s, expr->operand(1));
    if (expr->operand(2))
      deal_template_array(s, expr->operand(2));
    break;
    
  case ARRAY_REF:
    {
      Template_def * tdef =new Template_def;
      tdef->setsymbol(expr->symbol());
      fill_template_table(tdef, expr->operand(1));
      s->controlParent()->addAttribute(Temp_Def, (void *) tdef, sizeof(Template_def));
      expr->symbol()->addAttribute(Temp_Def, (void *) tdef, sizeof(Template_def));
    };
    break;

  default:
    printf("Wrong array index in declaration: %d\n", expr->id());
    break;

  };
}

void analyse_template(SgStatement * s)
{
  deal_template_array(s, s->expr(0));
}

void show_template(SgStatement * s)
{
  for (int i=0; i<s->numberOfAttributes(Temp_Def); i++) {
    ((Template_def *) ((s->getAttribute(i, Temp_Def))->getAttributeData()) ) ->show();
  };
}

extern SgType * gettype(int);

void add_template_declaration(SgStatement * s)
{
  for (int i=0; i<s->numberOfAttributes(Temp_Def); i++) {
    Template_def * tdef = (Template_def *) ((s->getAttribute(i, Temp_Def))->getAttributeData());

    SgSymbol * tempname = s->getSymbolInCurrentScope(mixstring, "rng", tdef->getsymbol()->identifier());

    tdef->setrng(tempname);

    SgValueExp * dim = new SgValueExp (tdef->getDim());
    
    SgExpression * range = new SgExpression(RANGE_OP, NULL, dim, NULL);

    SgExpression * ddot = new SgExpression(DDOT, range, NULL, NULL);

    SgExpression * index = new SgExpression(EXPR_LIST, ddot, NULL, NULL);

    SgExpression * array = new SgExpression(ARRAY_REF, index, NULL, tempname);

    SgType * stype = gettype(T_INT); 

    SgExpression * arraytype = new SgExpression(TYPE_OP, NULL, NULL, NULL, stype);

    SgStatement * decltemplate =new SgStatement(VAR_DECL, NULL, NULL, array, arraytype, NULL);
  
    s->childList1(0)->insertStmtBefore(*decltemplate, *s);

  };
}

//-----------------------;

void analyse_processor(SgStatement *);
void show_processor(SgStatement *);
void add_processor_declaration(SgStatement *);

void fill_table(Processor_def * pdef,SgExpression * expr)
{
  //printf("Current Expr: %d\n", expr->id());
  switch(expr->variant()) {
  case EXPR_LIST:
    fill_table(pdef, expr->operand(1));
    if (expr->operand(2))
      fill_table(pdef, expr->operand(2));
    break;

  case DDOT:
    fill_table(pdef, expr->operand(1));
    break;

  case RANGE_OP:
    {
      int start, end;
      if (expr->operand(1)) {

	if (expr->operand(1)->isInteger())
	  start=expr->operand(1)->valueInteger();
	else if (expr->operand(1)->symbol()->variant()==CONST_NAME) { // should be a parameter value ; 
	  //printf ("The id is %d\n", ((SgConstantSymb *) expr->operand(2)->symbol())->constantValue()->id() );
	  if (((SgConstantSymb *) expr->operand(1)->symbol())->constantValue()->isInteger())
	    start =((SgConstantSymb *) expr->operand(2)->symbol())->constantValue()->valueInteger();
	  else
	    printf("It should be an integer\n");
	}
	else
	  printf("Unknow processor boundaries.\n");
      }      
      else
	start=1;

      if (expr->operand(2)->isInteger())
	end=expr->operand(2)->valueInteger();
      else if (expr->operand(2)->symbol()->variant()==CONST_NAME) {  // should be a parameter value ; 
	//printf ("The id is %d\n", ((SgConstantSymb *) expr->operand(2)->symbol())->constantValue()->id() );
	if (((SgConstantSymb *) expr->operand(2)->symbol())->constantValue()->isInteger())
	  end =((SgConstantSymb *) expr->operand(2)->symbol())->constantValue()->valueInteger();
	else
	  printf("It should be an integer\n");
      }
      else 
	printf("Unknow processor boundaries.\n");
 
      pdef->add_dimention(start, end);
    };
    break;

  default:
    printf("Wrong array index in declaration: %d\n", expr->id());
  };
}

void deal_processor_array(SgStatement * s, SgExpression * expr)
{
  switch(expr->variant()) {
  case EXPR_LIST:
    deal_processor_array(s, expr->operand(1));
    if (expr->operand(2))
      deal_processor_array(s, expr->operand(2));
    break;
    
  case ARRAY_REF:
    {
      Processor_def * pdef =new Processor_def;
      pdef->setsymbol(expr->symbol());
      fill_table(pdef, expr->operand(1));
      s->controlParent()->addAttribute(Proc_Def, (void *) pdef, sizeof(Processor_def));
      expr->symbol()->addAttribute(Proc_Def, (void *) pdef, sizeof(Processor_def));
    };
    break;

  default:
    printf("Wrong array index in declaration: %d\n", expr->id());
    break;

  };
}

void analyse_processor(SgStatement * s)
{
  deal_processor_array(s, s->expr(0));
}

extern int getsizeof(char *);

void show_processor(SgStatement * s)
{
  for (int i=0; i<s->numberOfAttributes(Proc_Def); i++) {
    ((Processor_def *) ((s->getAttribute(i, Proc_Def))->getAttributeData()) ) ->show();
  };
  
}

void add_processor_declaration(SgStatement * s)
{
  for (int i=0; i<s->numberOfAttributes(Proc_Def); i++) {

    Processor_def * pdef = (Processor_def *) ((s->getAttribute(i, Proc_Def))->getAttributeData());

    // The following make up the processor shap declaration;
	 
    SgSymbol * proshp= s->getSymbolInCurrentScope(mixstring, "shp", pdef->getsymbol()->identifier());

    pdef->setshp(proshp);

    SgValueExp * dim = new SgValueExp (pdef->getDim());
    SgExpression * range = new SgExpression(RANGE_OP, NULL, dim, NULL);
    SgExpression * ddot = new SgExpression(DDOT, range, NULL, NULL);
    SgExpression * index = new SgExpression(EXPR_LIST, ddot, NULL, NULL);
    SgExpression * array = new SgExpression(ARRAY_REF, index, NULL, proshp);

    SgType * stype = gettype (T_INT); 
    SgExpression * arraytype = new SgExpression(TYPE_OP, NULL, NULL, NULL, stype);
  
    SgStatement * declshp = new SgStatement(VAR_DECL, NULL, NULL, array, arraytype, NULL);
  
    s->childList1(0)->insertStmtBefore(*declshp, *s);
  
    // the following is processor grid;
    
    SgSymbol * progrp= s->getSymbolInCurrentScope(mixstring, "grp", pdef->getsymbol()->identifier());

    pdef->setgrp(progrp);
    
    SgExpression * var = new SgExpression(VAR_REF, NULL, NULL, progrp);
    SgExpression * expr = new SgExpression(EXPR_LIST, var, NULL, NULL);
  
    SgType * gtype = gettype (T_INT); 
    SgExpression * exprtype = new SgExpression(TYPE_OP, NULL, NULL, NULL, gtype);

    SgStatement * declgrp =new SgStatement(VAR_DECL, NULL, NULL, expr, exprtype, NULL);
  
    s->childList1(0)->insertStmtBefore(*declgrp, *s);

  };
}

//------------------;

void caculate_template_local_size(SgStatement * s)
{
  //  show_template(s);

  for (int i=0; i<s->numberOfAttributes(Temp_Def); i++) {
    Template_def * tdef = (Template_def *) ((s->attributeValue(i, Temp_Def)));

    SgSymbol * tsymb = tdef->getsymbol();

    if (tsymb->numberOfAttributes(Dist_Def)!=0) {

      Distribute_def * ddef = (Distribute_def *) tsymb->getAttribute(0, Dist_Def)->getAttributeData();
    
	// dim record the real dimension distributed (except NOP)
      int dim=0;
      for (int j=0; j < tdef->getDim(); j++) {
      
	switch(ddef->distCode(j)) {
	case NOP:
	  tdef->setlocalsize(j, tdef->getEnd(j)-tdef->getStart(j)+1);
	  break;
	case CYCLIC:
	case BLOCK:
	  { 
	    if (ddef->gettosymbol()->numberOfAttributes(Proc_Def)) {
	      int temp;
	      int proc_number;

	      int start=((Processor_def *)(ddef->gettosymbol()->getAttribute(0,Proc_Def)->getAttributeData()))->getStart(dim);
	      int end=((Processor_def *)(ddef->gettosymbol()->getAttribute(0,Proc_Def)->getAttributeData()))->getEnd(dim);
	      proc_number=end-start+1;
	      ASSERT(proc_number != 0);

	      temp=(tdef->getEnd(j)-tdef->getStart(j)+1)/proc_number;

	      if ((tdef->getEnd(j)-tdef->getStart(j)+1)%proc_number)
		temp++;

	      tdef->setlocalsize(j, temp);
	
	    }
	    else
	      printf("Not distributed yet\n");
	  };
	  dim++;
	  break;
	default:
	  printf("no more code!\n");
	};
      };
    };
  };
}












