extern "C" {
#include <stdlib.h>
#include <stdio.h>
	   }

#include "compilerApi.h"

#include "anls_att.h"

extern char *mixstring;

Shadow_def::Shadow_def(Org tname)
{
  tag=tname;
  for (int i=0; i<MAX_SHADOW_DIM; i++){
    l[i]=0; 
    u[i]=0;
  }
}

Org Shadow_def::getTag()
{
  return tag;
}

void Shadow_def::setLower(int dim, int size)
{
  l[dim]=size;
}
  
void Shadow_def::setUpper(int dim, int size)
{
  u[dim]=size;
}

int Shadow_def::getLower(int dim)
{
  return l[dim];
}

int Shadow_def::getUpper(int dim)
{
  return u[dim];
}

void  Shadow_def::setScope(SgStatement *s)
{
  scope=s;
}

SgStatement *  Shadow_def::getScope()
{
  return scope;
}

void Shadow_def::show(void)
{

  printf ("tag: %d\n", tag);
  printf("lsize : ");
  for(int i=0; i<MAX_SHADOW_DIM;i++)
    printf("%3d", l[i]);
  printf("\n");

  printf("usize : ");
  for(int j=0; j<MAX_SHADOW_DIM;j++)
    printf("%3d", u[j]);
  printf("\n");

  printf("\n");
}

//--------------

void decide_global_shadow(SgStatement *s)
{ 

  for (int i=0; i<s->numberOfAttributes(Align_Def); i++) {
    Alignment_def * adef = ((Alignment_def *) ((s->getAttribute(i, Align_Def))->getAttributeData()) );    

    SgSymbol * array = s->getSymbolInCurrentScope(adef->getfromsymbol()->identifier());

    if (array->numberOfAttributes(Shad_Def)) {

      Shadow_def * dsdef = new Shadow_def (GNEEDED);
      
      
      for (int numOfDim=0; numOfDim < MAX_ALIGN_DIM; numOfDim++) {
	int lsize=0;
	int usize=0;
	
	for ( int numS=0; numS<array->numberOfAttributes(Shad_Def); numS++) {
	  Shadow_def * sdef = ((Shadow_def *) ((array->getAttribute(numS, Shad_Def))->getAttributeData()) );
	  
	  //sdef->show();
	  
	  if (sdef->getTag() == DETECTED) {
	    if (sdef->getLower(numOfDim) > lsize)
	      lsize = sdef->getLower(numOfDim);
	    if (sdef->getUpper(numOfDim) > usize)
	      usize = sdef->getUpper(numOfDim);
	  }
	  
	}
	
	dsdef->setLower(numOfDim, lsize);
	
	dsdef->setUpper(numOfDim, usize);
	
      }
      
      //      printf("here is global\n");
      //      dsdef->show();
      
      array->addAttribute(Shad_Def,(void *) dsdef, sizeof(Shadow_def));
    }
  }
}

//---------------------------------;


void get_shift_array(SgStatement *s)
{
  SgStatement * body;
  body = s->childList1(0);

  Forall_header_def * fhdef = (Forall_header_def *) s->getAttribute(0, FHDER_Def)->getAttributeData();

  for (int i=0; i<body->numberOfAttributes(Forall_Ref); i++) {
    forall_reference * fref = (forall_reference *) ((body->getAttribute(i, Forall_Ref))->getAttributeData());

    if (fref->getCommunication()==SHIFT) {
      int found =0;

      for ( int j=0; j<body->numberOfAttributes(SHIFTARRAY_Def); j++) {
	SgSymbol *array = (SgSymbol *) ((body->getAttribute(j, SHIFTARRAY_Def))->getAttributeData());
	
	if (array==fref->getexpression()->symbol()) {
	  found =1;
	  break;
	};
      };

      if (!found)
	body->addAttribute(SHIFTARRAY_Def, (void*)fref->getexpression()->symbol(), sizeof(SgSymbol)); 
      
    }
  }
}


void decide_local_shadow(SgStatement * s)
{

  SgStatement * body;
  body = s->childList1(0);

  for (int i=0; i<body->numberOfAttributes(SHIFTARRAY_Def); i++) {
    SgSymbol * array = (SgSymbol *) ((body->getAttribute(i, SHIFTARRAY_Def))->getAttributeData());

    if (array->numberOfAttributes(Shad_Def)) {

      Shadow_def * dsdef = new Shadow_def (LNEEDED);
      
      for (int numOfDim=0; numOfDim < MAX_ALIGN_DIM; numOfDim++) {
	int lsize=0;
	int usize=0;
	
	for ( int numS=0; numS<array->numberOfAttributes(Shad_Def); numS++) {
	  Shadow_def * sdef = ((Shadow_def *) ((array->getAttribute(numS, Shad_Def))->getAttributeData()) );
	  if (sdef->getTag()==DETECTED && sdef->getScope()==s) {
	    
	    if (sdef->getLower(numOfDim) > lsize)
	      lsize = sdef->getLower(numOfDim);
	    if (sdef->getUpper(numOfDim) > usize)
	      usize = sdef->getUpper(numOfDim);
	  }
	  
	}
      
	dsdef->setLower(numOfDim, lsize);
	
	dsdef->setUpper(numOfDim, usize);      
      }
      
      //dsdef->show();
      
      array->addAttribute(Shad_Def,(void *) dsdef, sizeof(Shadow_def));
    }
  }
}

void define_shadow_amount(SgStatement * s, int counter)
{
  SgStatement * body;
  body = s->childList1(0);

  SgSymbol * array = (SgSymbol *) ((body->getAttribute(counter, SHIFTARRAY_Def))->getAttributeData());

  //printf("define halo for %s\n", array->identifier());

  for (int i=0; i<array->numberOfAttributes(Shad_Def); i++) {
    Shadow_def * sdef = ((Shadow_def *) ((array->getAttribute(i, Shad_Def))->getAttributeData()) );
    if (sdef->getTag()==LNEEDED) {
      
      SgArrayType * arrayType=(SgArrayType *)array->type();
      int dim = arrayType->dimension();

      //printf("type : %d dim %d\n", arrayType->id(), dim);
      
      for(int j=0; j<dim; j++){

	SgValueExp * dimExp=new SgValueExp(j+1);

	SgValueExp * lowerExp= new SgValueExp(sdef->getLower(j));
	SgValueExp * upperExp= new SgValueExp(sdef->getUpper(j));

	SgSymbol * gtl=s->getSymbolInCurrentScope(mixstring, "gtl", array->identifier());
	SgSymbol * gtu=s->getSymbolInCurrentScope(mixstring, "gtu", array->identifier());
      
	SgExpression * range = new SgExpression(RANGE_OP, NULL, dimExp, NULL);
	SgExpression * ddot = new SgExpression(DDOT, range, NULL, NULL);
	SgExpression * index = new SgExpression(EXPR_LIST, ddot, NULL, NULL);

	SgExpression * gtlarray = new SgExpression(ARRAY_REF, index, NULL, gtl);
	SgExpression * gtuarray = new SgExpression(ARRAY_REF, index, NULL, gtu);

	SgAssignStmt * gtlAssign = new SgAssignStmt(*gtlarray, *lowerExp);
	SgAssignStmt * gtuAssign = new SgAssignStmt(*gtuarray, *upperExp);

	s->insertStmtBefore(*gtlAssign, *s->controlParent());
	s->insertStmtBefore(*gtuAssign, *s->controlParent());

      }

      break;
    }
  }
}


void add_write_halo(SgStatement * s, int counter)
{
  SgStatement * body;
  body = s->childList1(0);

  SgSymbol * array = (SgSymbol *) ((body->getAttribute(counter, SHIFTARRAY_Def))->getAttributeData());

  //printf("write halo for %s\n", array->identifier());

  SgSymbol * dad=s->getSymbolInCurrentScope(mixstring, "dad", array->identifier());  
  SgExpression * dadExp = new SgExpression(VAR_REF, NULL, NULL, dad);
  
  SgSymbol * gtl=s->getSymbolInCurrentScope(mixstring, "gtl", array->identifier());
  SgExpression * gtlarray = new SgExpression(ARRAY_REF, NULL, NULL, gtl);

  SgSymbol * gtu=s->getSymbolInCurrentScope(mixstring, "gtu", array->identifier());
  SgExpression * gtuarray = new SgExpression(ARRAY_REF, NULL, NULL, gtu);
  
  SgSymbol * writeHalo=s->getSymbolInCurrentScope(mixstring, "write_halo", NULL);
  SgCallStmt * func = new SgCallStmt(*writeHalo);

  func->addArg(*dadExp);
  func->addArg(*gtlarray);
  func->addArg(*gtuarray);
  
  s->insertStmtBefore(*func, *s->controlParent());

}


void write_halo(SgStatement *s)
{
  SgStatement * body;
  body = s->childList1(0);

  for (int i=0; i<body->numberOfAttributes(SHIFTARRAY_Def); i++) {
    SgSymbol * array = (SgSymbol *) ((body->getAttribute(i, SHIFTARRAY_Def))->getAttributeData());

    if (array->numberOfAttributes(Shad_Def)) {

      define_shadow_amount(s, i);
      add_write_halo(s, i);

    }
  }
}


void check_ghost(SgStatement * s)
{

  get_shift_array(s);

  decide_local_shadow(s);

  write_halo(s);

}


