// 	pass.C
//	  pass of transform data mapping statements to uniform 
//
//
// written by Zheng Gengbin

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#if STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#else
#include <strings.h>
#endif

#include        <stdio.h>
#include        <malloc.h>

#include "compilerApi.h"
#include "pass.h"

#include "ext_func.h"

int num_procs=NUMBER_OF_PROC;

char * mixstring="pcrc";

Flag in_main_program = off;
Flag in_procedure = off;

int proc_info_collected=0;
int dist_info_collected = 0;
int align_info_collected = 0;
int temp_info_collected = 0;
int alloc_info_collected=0;

int do_level = 0;

PassType passtype;

SgFile * global_file;

//
extern void collapse_aligns(SgStatement *);
//
extern void align_coef(SgStatement *);
//
extern void complete_distribute_elements(SgStatement *);

char *newOutputFileName(char *inputFileName)
{
  char  *out;
  int   nameLength = strlen(inputFileName);
  
  out=(char *)malloc((unsigned)(nameLength+5));

  while (inputFileName[nameLength-1]!='.') {
    nameLength --;
  }

  for ( int i = 0; i<nameLength; i++) {
    out[i]=inputFileName[i];
  }

  out[nameLength-1]='\0';

  out=strcat(out,"_1.dep");

  return(out);
}

void analyse(SgStatement * s)
{
  switch(s->variant()) {

  case GLOBAL:
    {
      int org_numberOfChildren=s->numberOfChildrenList1();
      SgStatement * * children = new SgStatement * [org_numberOfChildren];

      int i;

      for (i=0; i<org_numberOfChildren; i++)
	children[i]=s->childList1(i);
      for (i=0; i<org_numberOfChildren; i++)
	analyse(children[i]);
      
      delete [] children;

      break;
    };

  case PROG_HEDR:
    {
      int org_numberOfChildren=s->numberOfChildrenList1();
      SgStatement * * children = new SgStatement * [org_numberOfChildren];

      int i; 

      for (i=0; i<org_numberOfChildren; i++)
	children[i]=s->childList1(i);

      // do nothing?;
      in_main_program=on;

	// do some initialization 
//      clear_proc_array();

      for (i=0; i<org_numberOfChildren; i++)
	analyse(children[i]);
      
      delete [] children;

      in_main_program=off;
	
      if (passtype == ALIGN_COLLAPSE_PASS)
      {
	collapse_aligns(s);
      }
      if (passtype == ALIGN_ROOT_TEMP_PASS)
      {
	collect_align_root(s);
	is_align_root_template(s);
      }
      if (passtype == ALIGN_ROOT_DIST_PASS)
      {
	collect_align_root(s);
	display_align_root(s);
	is_align_root_distribute(s);
      }


      break;

    };

  case INTERFACE_STMT:
    {
      SgProcHedrStmt * prochedr = (SgProcHedrStmt *)s;
      int org_numberOfChildren=s->numberOfChildrenList1();
      SgStatement * * children = new SgStatement * [org_numberOfChildren];

      int i;

      for (i=0; i<org_numberOfChildren; i++)
	children[i]=s->childList1(i);

      for (i=0; i<org_numberOfChildren; i++)
	analyse(children[i]);
      
      delete [] children;

      break;
    }

  case PROC_HEDR:
    {
      SgProcHedrStmt * prochedr = (SgProcHedrStmt *)s;
      int org_numberOfChildren=s->numberOfChildrenList1();
      SgStatement * * children = new SgStatement * [org_numberOfChildren];

      int i;

      for (i=0; i<org_numberOfChildren; i++)
	children[i]=s->childList1(i);

      in_procedure=on;

      for (i=0; i<prochedr->numberOfParameters(); i++) {
	prochedr->parameter(i)->addAttribute(DUMP_Def, NULL, 0);
	prochedr->addAttribute(DUMP_Def, NULL, 0);
      }

	// do some initialization 
//      clear_proc_array();

      for (i=0; i<org_numberOfChildren; i++)
	analyse(children[i]);
      
      delete [] children;
      
      in_procedure = off;

      if (passtype == ALIGN_COLLAPSE_PASS)
      {
	collapse_aligns(s);
      }
      if (passtype == ALIGN_ROOT_TEMP_PASS)
      {
	collect_align_root(s);
	is_align_root_template(s);
      }
      if (passtype == ALIGN_ROOT_DIST_PASS)
      {
	collect_align_root(s);
	display_align_root(s);
	is_align_root_distribute(s);
      }

      break;
    };

  case FUNC_HEDR:
    {
      SgFuncHedrStmt * prochedr = (SgFuncHedrStmt *)s;
      int org_numberOfChildren=s->numberOfChildrenList1();
      SgStatement * * children = new SgStatement * [org_numberOfChildren];

      int i;

      for (i=0; i<org_numberOfChildren; i++)
	children[i]=s->childList1(i);

      in_procedure=on;

      for (i=0; i<prochedr->numberOfParameters(); i++) {
	prochedr->parameter(i)->addAttribute(DUMP_Def, NULL, 0);
	prochedr->addAttribute(DUMP_Def, NULL, 0);
      }

      for (i=0; i<org_numberOfChildren; i++)
	analyse(children[i]);
      
      delete [] children;
      
      in_procedure = off;

      if (passtype == ALIGN_COLLAPSE_PASS)
      {
	collapse_aligns(s);
      }
      if (passtype == ALIGN_ROOT_TEMP_PASS)
      {
	collect_align_root(s);
	is_align_root_template(s);
      }
      if (passtype == ALIGN_ROOT_DIST_PASS)
      {
	collect_align_root(s);
	display_align_root(s);
	is_align_root_distribute(s);
      }

      break;
    };

  case COMBINED_DIRECTIVE_DECL:
	if (passtype == COMBINE_PASS)
		transform_combined(s);
    break;

  case PROCESSORS_STMT:
	if (passtype == COLLECT_PROCESSOR_INFO && proc_info_collected==0)
		collect_proc_attributes(s);
    break;

  case TEMPLATE_STMT:
	if (passtype == COLLECT_ALIGN_DIST_INFO && temp_info_collected==0)
		collect_temp_attributes(s);
    break;

  case ALIGN_DECL:
	if (passtype == ALIGN_COLON_PASS)
		trans_align_colon_to_dummy(s);
	if (passtype == COLLECT_ALIGN_DIST_INFO && align_info_collected==0)
		collect_align_attributes(s);
	if (passtype == ALIGN_COEF_PASS)
		align_coef(s);
    break;
    
  case DISTRIBUTE_DECL:
//    analyse_distribute(s);
	if (passtype == DIST_COMP_PASS)
		complete_distribute_elements(s);
	if (passtype == COLLECT_ALIGN_DIST_INFO && dist_info_collected == 0)
		collect_dist_attributes(s);
    break;

  case INHERIT_DECL:
//    analyse_inherit(s);
    break;

  case VAR_DECL:
    break;

  case ALLOCATE_STMT:
	if (passtype == COLLECT_ALIGN_DIST_INFO && alloc_info_collected==0)
		collect_alloc_attributes(s);
    break;

  case FOR_NODE:
    {
      do_level++;

      int org_numberOfChildren=s->numberOfChildrenList1();

      SgStatement * * children = new SgStatement * [org_numberOfChildren];

      int i;

      for (i=0; i<org_numberOfChildren; i++)
	children[i]=s->childList1(i);

      for (i=0; i<org_numberOfChildren; i++)
	analyse(children[i]);
      
      delete [] children;
      
      do_level--;

      break;
    }

  case LOGIF_NODE:
  case ELSEIF_NODE:
  case LOGIF_STMT:
  case SWITCH_NODE:
  case CASE_NODE:
    {
      int org_numberOfChildren1=s->numberOfChildrenList1();

      SgStatement * * children1 = new SgStatement * [org_numberOfChildren1];

      int org_numberOfChildren2=s->numberOfChildrenList2();

      SgStatement * * children2 = new SgStatement * [org_numberOfChildren2];

      int i;

      for (i=0; i<org_numberOfChildren1; i++)
	children1[i]=s->childList1(i);

      for (i=0; i<org_numberOfChildren1; i++)
	analyse(children1[i]);
      
      delete [] children1;

      for (i=0; i<org_numberOfChildren2; i++)
	children2[i]=s->childList2(i);			// zheng.gb

      for (i=0; i<org_numberOfChildren2; i++)
	analyse(children2[i]);
      
      delete [] children2;

      break;

    }

  case ASSGOTO_NODE:
  case STOP_STAT:
    // do nothing;
    break;


  case PROC_STAT:
    //do nothing at presenttime. we will analyse INTERFACE later;
    break;

  case ASSIGN_STAT:
//    analyse_assign(s);
    break;

  case FORALL_STMT:
    {       
/*
      analyse_forall_header(s);
      get_forall_reference(s);

      Forall_header_def * fhdef=(Forall_header_def *) s->getAttribute(0, FHDER_Def)->getAttributeData();
    
      for (int i=0; i<fhdef->getindexnumber(); i++) {
	SgSymbol * index=fhdef->getindexsymb(i);
	index_forall_reference(i, index, s->childList1(0));
      };

      detect_forall_reference(s);
*/
      break;
    }

  case PRINT_STAT:
  case WRITE_STAT:
//    analyse_datatransfer(s);
    break;

  case CONTROL_END:
  case PARAM_DECL:
  case ENDDO_STMT:
  case  FORMAT_STAT:
   // statement need do nothing;
    break;

  default:
//    printf("Not touched statement in analyse on line %d\n", s->lineNumber());
    break;
  };

}

static void
usage()
{
        printf("Usage: stdpass [-nabcdh] HPF_file.dep \n");
        printf("       -h    : usage          \n");
        printf("       -n    : specify the number of processors . \n");
        printf("       -a    : Covert combined directive to statements.\n");
        printf("       -b    : Align statement to dummy uniform .\n");
        printf("       -c    : Collapse align statement to root.\n");
        printf("       -d    : Convert align spec to a*x+b uniform .\n");
}

main(int argc, char * argv[]) 
{
  SgProject * project;
  char *file_in , *file_out;
  PassType   pt[10];
  int no,i,j;

	argv++;
        argc--;
	no=0;
        while ((argc > 0) && ((*argv)[0] == '-') )
        {
                i=1;
                while (argv[0][i] != '\0') {
			switch (argv[0][i]) {
			case 'a'  :
				pt[no++]=COMBINE_PASS; break;
			case 'b'  :
				pt[no++]=COLLECT_PROCESSOR_INFO; break;
			case 'c'  :
				pt[no++]=ALIGN_COLON_PASS; break;
			case 'd'  :
				pt[no++]=DIST_COMP_PASS; break;
			case 'e'  :
				pt[no++]=COLLECT_ALIGN_DIST_INFO; break;
			case 'f'  :
				pt[no++]=ALIGN_ROOT_TEMP_PASS; break;
			case 'g'  :
				pt[no++]=ALIGN_COLLAPSE_PASS; break;
			case 'h'  :
				pt[no++]=ALIGN_COEF_PASS; break;
			case 'i'  :
				pt[no++]=ALIGN_ROOT_DIST_PASS; break;
			case 'n'  :
				if (argv[0][i+1]!='\0') {
					num_procs = atoi(argv[0]+i+1);
                                        i=strlen(argv[0]);
                                        continue;
                                }
                                else {
                                   if (argc==1) {
                                        printf("warning : Unknown option \"n\" ignored.\n");
                                   }
                                   else {
                                        argc--;
                                        argv++;
					num_procs = atoi(argv[0]);
                                        i=strlen(argv[0])-1;   /* quit loop */
                                   }
                                }
                                break;
			default:
                                printf("warning 422: Unknown option \"%c\" ignored.\n",argv[0][i]);
				usage() ; exit(1);
			}
			i++;
		}
		argc--;
		argv++;
	}
	// if not specify any pass type , suggest a reasonable one.
	if (no == 0)
	{
		pt[no++]=COMBINE_PASS; 
		pt[no++]=COLLECT_PROCESSOR_INFO; 
		pt[no++]=ALIGN_COLON_PASS; 
		pt[no++]=DIST_COMP_PASS; 
		pt[no++]=COLLECT_ALIGN_DIST_INFO; 
		pt[no++]=ALIGN_ROOT_TEMP_PASS; 
		pt[no++]=ALIGN_COLLAPSE_PASS; 
		pt[no++]=ALIGN_COEF_PASS; 
		pt[no++]=ALIGN_ROOT_DIST_PASS; 
	}

  argc++; argv--;
  project = new SgProject("test", argc, argv);

  for (i=0; i< project->numberOfFiles(); i++) {

    global_file = &(project->file(i));
    file_in=global_file->filept->filename;

    SgStatement * s;

    char comment[100];
    for (j=0; j<no ; j++)
    {
	passtype=pt[j];

	switch (passtype) {
	case COMBINE_PASS :
		sprintf(comment,"%s","Combine directive");
		break;
	case ALIGN_COLON_PASS :
		sprintf(comment,"%s","Align to DUMMY uniform");
		break;
	case DIST_COMP_PASS :
		sprintf(comment,"%s","Complete distribute statement .");
		break;
	case COLLECT_ALIGN_DIST_INFO :
		sprintf(comment,"%s","Collect align and distribute statement in symbol .");
		break;
	case ALIGN_ROOT_TEMP_PASS :
		sprintf(comment,"%s","Make sure every align root is template.");
		break;
	case ALIGN_COLLAPSE_PASS :
		sprintf(comment,"%s","Align Collpase to root ");
		break;
	case ALIGN_COEF_PASS :
		sprintf(comment,"%s","Align spec convert to a*X+b uniform");
		break;
	case ALIGN_ROOT_DIST_PASS :
		sprintf(comment,"%s","Make sure every align root is distributed.");
		break;
	}
	printf("***********  PASS %s **********\n",comment);
    	s = global_file->firstStatement();
    	analyse(s);

	if (passtype == COLLECT_ALIGN_DIST_INFO)
		align_info_collected=dist_info_collected=temp_info_collected=alloc_info_collected=1;
	if (passtype == COLLECT_PROCESSOR_INFO)
		proc_info_collected=1;
    }

    file_out=newOutputFileName(file_in);
    global_file->saveDepFile(file_out);
    
  };

  return 0;
}
