
#include <math.h>
#include <time.h>
#include <stdio.h>
#include <stdlib.h>

#include "generator.h"

void main(int argc, char *argv[]) {

   int rank,t_rank,p_rank;
   unsigned int seed;

   srand((unsigned int) time(NULL));

   header();
   p_rank = processor();
   rank = array(p_rank);
   t_rank = template(rank,p_rank);
   distribute(t_rank,p_rank);
   align(rank);
   statements(rank);
   done(rank);

}

header() {
   printf("      PROGRAM LEVEL4\n");
}

array(int p_rank) {

   int i,j,rank, variation;
   char comma[MOST];

   rank = p_rank + rand()%3;   /* the rank of arrays >= p_rank */
   num_of_arrays = LEAST + rand()%(MOST-LEAST);
   for (i=0; i<num_of_arrays; i++)
      comma[i] = ',';
   comma[num_of_arrays-1] = '\n';

   for (i=0; i<num_of_arrays; i++) {
      arrays[i].name = array_name[i];
      arrays[i].rank = rank;
      for (j=0; j<arrays[i].rank; j++) {
         variation = 5*(rand()%VARIANT);  /* make it multiple of 5 */
         variation = (rand()%2)?variation:-variation;
         arrays[i].extent[j] = EXTENT + variation;
         arrays[i].stride[j] = rand()%MAX_STRIDE + 1;
         arrays[i].offset[j] = rand()%MAX_OFFSET;
      }
      arrays[i].template = i;
   }
   
   printf("      REAL ");
   for (i=0; i<num_of_arrays; i++) {
      printf("%s(",arrays[i].name);
      for (j=0; j<arrays[i].rank-1; j++)
         printf("%d,",arrays[i].extent[j]);
      printf("%d)%c ",arrays[i].extent[arrays[i].rank-1],comma[i]);
   }
   printf("\n");
   return rank;
}

processor() {

   int dim,t,p1,p2,p3;
   dim = rand()%3 + 1;
   switch (dim) {
      case 1:
         t = rand() % 3;
         /*  p1 = pow(2,t+1); the number of processors: 2, 4, 8 */
         p1 = 2;
         while (t != 0) { p1 = p1*2; t--;};
         printf("!HPF$ PROCESSORS P(%d)\n",p1);
         break;
      case 2: /* (2,2), (2,4),(4,2) cases */
         p1 = (rand()%2)?2:4;
         if (p1 == 4)
            p2 = 2;
         else
            p2 = (rand()%2)?2:4;
         printf("!HPF$ PROCESSORS P(%d,%d)\n",p1,p2);
         break;
      case 3:
         printf("!HPF$ PROCESSORS P(2,2,2)\n");
         break;
      default:
         break;
   }
   return dim;
}

template(int rank, int p_rank) {

   int i,j,t_rank, variation;
   char comma[MOST];

   num_of_temps = num_of_arrays;     /* let's have 1-1 onto for now */
   for (i=0; i<num_of_temps; i++)
      comma[i] = ',';
   comma[num_of_temps-1] = '\n';

   t_rank = rank;

   for (i=0; i<num_of_arrays; i++) {
      templates[i].name = temps_name[i];
      templates[i].rank = t_rank;
      for (j=0; j<templates[i].rank; j++) {
         templates[i].extent[j] = (arrays[i].stride[j])*(arrays[i].extent[j])
                                + arrays[i].offset[j] + 5*(rand()%VARIANT);
      }
   }
   
   printf("!HPF$ TEMPLATE ");
   for (i=0; i<num_of_temps; i++) {
      printf("%s(",templates[i].name);
      for (j=0; j<templates[i].rank-1; j++)
         printf("%d,",templates[i].extent[j]);
      printf("%d)%c",templates[i].extent[templates[i].rank-1],comma[i]);
   }
   printf("\n");
   return t_rank;
};

distribute(int t_rank,int p_rank) {
   
   int i,j,hit,rank_dif = t_rank-p_rank;

   for (i=0; i<num_of_temps; i++) {
      for (j=0; j<templates[i].rank; j++) {
         if (rand()%2)
            templates[i].dist[j] = "BLOCK";
         else
            templates[i].dist[j] = "CYCLIC";
      }
      for (j=0; j<rank_dif; j++) {
         hit = rand()%t_rank;
         while (templates[i].dist[hit] == "*") hit = rand()%t_rank;
         templates[i].dist[hit] = "*";
      }
   }

   for (i=0; i<num_of_temps; i++) {
      printf("!HPF$ DISTRIBUTE ");
      printf("%s(",templates[i].name);
      for (j=0; j<templates[i].rank-1; j++)
         printf("%s,",templates[i].dist[j]);
      printf("%s) ONTO P \n",templates[i].dist[templates[i].rank-1]);
   }
   return 0;
};

align(int rank) {

   int i,j;

   for (i=0; i<num_of_arrays; i++) {
      printf("!HPF$ ALIGN ");
      printf("%s(",arrays[i].name);
      for (j=0; j<arrays[i].rank-1; j++)
         printf("i%d,",j);
      printf("i%d) WITH %s(",rank-1,templates[i].name);
      for (j=0; j<arrays[i].rank-1; j++)
         printf("%d*i%d+%d,",arrays[i].stride[j],j,arrays[i].offset[j]);
      printf("%d*i%d+%d) \n",
              arrays[i].stride[rank-1],rank-1,arrays[i].offset[rank-1]);
   }
   return 0;
};

statements(int rank) {

   int i,j,k,num_of_stmts,num_of_terms,next_lhs=0;
   int min_extent,true_extent,max_stride,low,up,step;
   int L[TERMS+1][7],U[TERMS+1][7],S[TERMS+1][7],l,u,s;
   char operator[TERMS+1], comma[MOST];

   for (i=0; i<num_of_arrays; i++)
      comma[i] = ';';
   comma[num_of_arrays-1] = '\n';

   printf("      ");
   for (i=0; i<num_of_arrays; i++)
      printf("%s = %d%c ",arrays[i].name,i,comma[i]);

   for (i=0; i<rank; i++)
      comma[i] = ',';
   comma[rank-1] = ')';

   num_of_stmts = rand()%MOST + 1;      /* number of statements to be gen. */
   for (i=0; i<num_of_stmts; i++) {
      num_of_terms = rand()%TERMS + 2;  /* at least 2, including lhs */
      operator[0] = '=';
      for (k=1; k<num_of_terms; k++) {
         operator[k] = (rand()%2)?'+':'-';
      }
      operator[num_of_terms-1] = '\n';

      for (j=0; j<rank; j++) {
         /* analyze extent constraint in the dimension */
         min_extent = BIG;
         for (k=0; k<num_of_terms; k++) {
            min_extent = 
               min(min_extent,arrays[(next_lhs+k)%num_of_arrays].extent[j]);
         }
         max_stride = rand()%MAX_STRIDE + 1;
         true_extent = min_extent/max_stride;
         if (true_extent <= 0) {
            printf ("bad luck ! rerun the program \n");
            return;
         }
         
         /* using true_extent and max_stride, 
            generate triplets for this dimension */ 
         for (k=0; k<num_of_terms; k++) {
            s = rand()%max_stride + 1;
            l = rand()%MAX_OFFSET + 1;
            u = (true_extent-1)*s + l; 
            if (u > arrays[(next_lhs+k)%num_of_arrays].extent[j]) {
               l = 1;
               u = (true_extent-1)*s + l;
            }
            L[k][j] = l;
            U[k][j] = u; 
            S[k][j] = s;
         }
      }
      /* output the statement */
      printf("     ");
      for (k=0; k<num_of_terms; k++) {
         printf("%s(",arrays[(next_lhs+k)%num_of_arrays].name);
         for (j=0; j<rank; j++) {
         /* we distinguish 8 representations for array sections */
            low = (L[k][j] == 1)?4:0;
             up = (U[k][j] == arrays[(next_lhs+k)%num_of_arrays].extent[j])?2:0;
            step = (S[k][j] == 1)?1:0;
            switch (low | up | step) {
               case 0:
                  printf("%d:%d:%d%c",L[k][j],U[k][j],S[k][j],comma[j]);
                  break; 
               case 1:
                  printf("%d:%d%c",L[k][j],U[k][j],comma[j]);
                  break; 
               case 2:
                  printf("%d::%d%c",L[k][j],S[k][j],comma[j]);
                  break; 
               case 3:
                  printf("%d:%c",L[k][j],comma[j]);
                  break; 
               case 4:
                  printf(":%d:%d%c",U[k][j],S[k][j],comma[j]);
                  break; 
               case 5:
                  printf(":%d%c",U[k][j],comma[j]);
                  break; 
               case 6:
                  printf("::%d%c",S[k][j],comma[j]);
                  break; 
               case 7:
                  printf(":%c",comma[j]);
                  break; 
               default:
                  break;
            }
         }
         printf(" %c ",operator[k]);
      }
      next_lhs = (next_lhs+num_of_terms)%num_of_arrays;
   }

};

done(int rank) {
   char idx[7] = {'i','1','1','1','1','1','1'};
   char comma[7];
   int i;
 
   for (i=0; i<rank; i++)
      comma[i] = ',';
   comma[rank-1] = ')';

   printf("      PRINT *, (X0(");
   for (i=0; i<rank; i++)
      printf("%c%c",idx[i],comma[i]);
   printf(", i = 1, 10)\n");
   printf("      END\n");
}
